import platform
import socket

import pytest

skip_on_windows = pytest.mark.skipif(
    platform.system() == "Windows", reason="Test is broken skipped on Windows"
)


def get_available_port():
    with socket.socket() as s:
        s.bind(("", 0))
        return s.getsockname()[1]



from __future__ import annotations

import uuid
from datetime import datetime, timezone
from typing import Annotated, Any

from pydantic import (
    AfterValidator,
    BaseModel,
    Field,
    ValidationInfo,
    computed_field,
    field_validator,
)

import ragna.core


class User(BaseModel):
    name: str
    data: dict[str, Any] = Field(default_factory=dict)


class ApiKeyCreation(BaseModel):
    name: str
    expires_at: datetime


class ApiKey(BaseModel):
    id: uuid.UUID = Field(default_factory=uuid.uuid4)
    name: str
    expires_at: datetime
    obfuscated: bool = True
    value: str

    @field_validator("expires_at")
    @classmethod
    def _set_utc_timezone(cls, v: datetime) -> datetime:
        if v.tzinfo is None:
            return v.replace(tzinfo=timezone.utc)
        else:
            return v.astimezone(timezone.utc)

    @computed_field  # type: ignore[misc]
    @property
    def expired(self) -> bool:
        return datetime.now(timezone.utc) >= self.expires_at

    @field_validator("value")
    @classmethod
    def _maybe_obfuscate(cls, v: str, info: ValidationInfo) -> str:
        if not info.data["obfuscated"]:
            return v

        i = min(len(v) // 6, 3)
        if i > 0:
            return f"{v[:i]}***{v[-i:]}"
        else:
            return "***"


def _set_utc_timezone(v: datetime) -> datetime:
    if v.tzinfo is None:
        return v.replace(tzinfo=timezone.utc)
    else:
        return v.astimezone(timezone.utc)


UtcDateTime = Annotated[datetime, AfterValidator(_set_utc_timezone)]


class Components(BaseModel):
    documents: list[str]
    source_storages: list[dict[str, Any]]
    assistants: list[dict[str, Any]]


class DocumentRegistration(BaseModel):
    name: str
    metadata: dict[str, Any] = Field(default_factory=dict)


class Document(BaseModel):
    id: uuid.UUID = Field(default_factory=uuid.uuid4)
    name: str
    metadata: dict[str, Any]


class Source(BaseModel):
    # See orm.Source on why this is not a UUID
    id: str
    document: Document
    location: str
    content: str
    num_tokens: int


class Message(BaseModel):
    id: uuid.UUID = Field(default_factory=uuid.uuid4)
    content: str
    role: ragna.core.MessageRole
    sources: list[Source] = Field(default_factory=list)
    timestamp: UtcDateTime = Field(default_factory=lambda: datetime.now(timezone.utc))


class ChatCreation(BaseModel):
    name: str
    document_ids: list[uuid.UUID]
    source_storage: str
    assistant: str
    params: dict[str, Any] = Field(default_factory=dict)


class Chat(BaseModel):
    id: uuid.UUID = Field(default_factory=uuid.uuid4)
    name: str
    documents: list[Document]
    source_storage: str
    assistant: str
    params: dict[str, Any]
    messages: list[Message] = Field(default_factory=list)
    prepared: bool = False



import uuid
from datetime import datetime

import emoji
import panel as pn
import param

from ragna.deploy import _schemas as schemas
from ragna.deploy._engine import Engine


class ApiWrapper(param.Parameterized):
    def __init__(self, engine: Engine):
        super().__init__()
        self._user = pn.state.user
        self._engine = engine

    async def get_chats(self):
        json_data = [
            chat.model_dump(mode="json")
            for chat in self._engine.get_chats(user=self._user)
        ]
        for chat in json_data:
            chat["messages"] = [self.improve_message(msg) for msg in chat["messages"]]
        return json_data

    async def answer(self, chat_id, prompt):
        async for message in self._engine.answer_stream(
            user=self._user, chat_id=uuid.UUID(chat_id), prompt=prompt
        ):
            yield self.improve_message(message.model_dump(mode="json"))

    def get_components(self):
        return self._engine.get_components()

    async def start_and_prepare(
        self, name, documents, source_storage, assistant, params
    ):
        chat = self._engine.create_chat(
            user=self._user,
            chat_creation=schemas.ChatCreation(
                name=name,
                document_ids=[document.id for document in documents],
                source_storage=source_storage,
                assistant=assistant,
                params=params,
            ),
        )
        await self._engine.prepare_chat(user=self._user, id=chat.id)
        return str(chat.id)

    def improve_message(self, msg):
        msg["timestamp"] = datetime.strptime(msg["timestamp"], "%Y-%m-%dT%H:%M:%S.%fZ")
        msg["content"] = emoji.emojize(msg["content"], language="alias")
        return msg



"""
UI Helpers
"""

from typing import Any, Dict, Iterable, Union

import panel as pn

"""
the structure of css_modifiers is as follows:
{
    "directory name under css/":[list of panel classes that will be modified],
    ...
}

Each CSS modifier file that needs to be loaded, is infered from the panel class name and the directory name.
For example, with value :
{ "foobar":[pn.widgets.TextInput] }
the css file that will be loaded is css/foobar/textinput.css

"""

css_modifiers = {
    "global": [
        pn.widgets.TextInput,
        pn.widgets.Select,
        pn.widgets.Button,
        pn.layout.Divider,
    ],
    "source_accordion": [
        pn.layout.Accordion,
        pn.layout.Card,
        pn.pane.HTML,
        pn.pane.Markdown,
    ],
    "chat_info": [pn.pane.Markdown, pn.widgets.Button],
    "central_view": [pn.Column, pn.Row, pn.pane.HTML],
    "chat_interface": [
        pn.widgets.TextInput,
        pn.pane.Markdown,
        pn.widgets.button.Button,
        pn.Column,
        pn.chat.ChatInterface,
    ],
    "right_sidebar": [pn.widgets.Button, pn.Column, pn.pane.Markdown],
    "left_sidebar": [pn.widgets.Button, pn.pane.HTML, pn.Column],
    "main_page": [pn.Row],
    "modal_configuration": [
        pn.widgets.IntSlider,
        pn.layout.Card,
        pn.Row,
        pn.widgets.Button,
        pn.widgets.FileInput,
    ],
}


def apply_design_modifiers():
    css_filepaths = []
    for dir, classes in css_modifiers.items():
        for cls in classes:
            css_filepath = f"css/{dir}/{cls.__name__.lower()}.css"
            add_modifier(cls, css_filepath)
            css_filepaths.append(css_filepath)

    return css_filepaths


def add_modifier(
    modifier_class: pn.viewable.Viewable,
    modifications: Dict[str, Any],
    property: str = "stylesheets",
):
    properties = pn.theme.fast.Fast.modifiers.setdefault(modifier_class, {})
    property_modifications = properties.setdefault(property, [])
    property_modifications.append(modifications)


"""
CSS helpers, constants and UI Helpers
"""


def divider():
    return pn.layout.Divider(css_classes=["default_divider"])


def css(selector: Union[str, Iterable[str]], declarations: dict[str, str]) -> str:
    return "\n".join(
        [
            f"{selector if isinstance(selector, str) else ', '.join(selector)} {{",
            *[f"    {property}: {value};" for property, value in declarations.items()],
            "}",
        ]
    )


MAIN_COLOR = "#DF5538"  # "rgba(223, 85, 56, 1)"

# set modal height
CONFIG_MODAL_MIN_HEIGHT = 610
CONFIG_MODAL_MAX_HEIGHT = 850
CONFIG_MODAL_WIDTH = 800


CSS_VARS = css(
    ":root",
    {
        "--body-font": "'Inter', sans-serif !important",
        "--accent-color": f"{MAIN_COLOR} !important",
    },
)


message_loading_indicator = f""" 
            <div style="height:48px;">
                <svg xmlns="http://www.w3.org/2000/svg" width="48" height="48" viewBox="0 0 24 24">
                    <circle cx="18" cy="12" r="0" fill="{MAIN_COLOR}">
                        <animate attributeName="r" begin=".67" calcMode="spline" dur="1.5s" keySplines="0.2 0.2 0.4 0.8;0.2 0.2 0.4 0.8;0.2 0.2 0.4 0.8" repeatCount="indefinite" values="0;2;0;0"/>
                    </circle>
                    <circle cx="12" cy="12" r="0" fill="{MAIN_COLOR}">
                        <animate attributeName="r" begin=".33" calcMode="spline" dur="1.5s" keySplines="0.2 0.2 0.4 0.8;0.2 0.2 0.4 0.8;0.2 0.2 0.4 0.8" repeatCount="indefinite" values="0;2;0;0"/>
                    </circle>
                    <circle cx="6" cy="12" r="0" fill="{MAIN_COLOR}">
                        <animate attributeName="r" begin="0" calcMode="spline" dur="1.5s" keySplines="0.2 0.2 0.4 0.8;0.2 0.2 0.4 0.8;0.2 0.2 0.4 0.8" repeatCount="indefinite" values="0;2;0;0"/>
                    </circle>
                </svg>
            </div>
            """



import pytest

import ragna


@pytest.fixture(autouse=True)
def tmp_local_root(tmp_path):
    old = ragna.local_root()
    try:
        yield ragna.local_root(tmp_path)
    finally:
        ragna.local_root(old)



import pytest

from ragna import Rag, assistants, source_storages
from ragna.core import Assistant, LocalDocument, RagnaException


@pytest.fixture()
def demo_document(tmp_path, request):
    path = tmp_path / "demo_document.txt"
    with open(path, "w") as file:
        file.write(f"{request.node.name}\n")
    return LocalDocument.from_path(path)


class TestChat:
    def chat(
        self,
        documents,
        source_storage=source_storages.RagnaDemoSourceStorage,
        assistant=assistants.RagnaDemoAssistant,
        **params,
    ):
        return Rag().chat(
            documents=documents,
            source_storage=source_storage,
            assistant=assistant,
            **params,
        )

    def test_params_validation_unknown(self, demo_document):
        params = {
            "bool_param": False,
            "int_param": 1,
            "float_param": 0.5,
            "string_param": "arbitrary_value",
        }
        with pytest.raises(RagnaException, match="unknown") as exc_info:
            self.chat(documents=[demo_document], **params)

        msg = str(exc_info.value)
        for param, value in params.items():
            assert f"{param}={value!r}" in msg

    def test_params_validation_missing(self, demo_document):
        class ValidationAssistant(Assistant):
            def answer(
                self,
                messages,
                bool_param: bool,
                int_param: int,
                float_param: float,
                string_param: str,
            ):
                pass

        with pytest.raises(RagnaException, match="missing") as exc_info:
            self.chat(documents=[demo_document], assistant=ValidationAssistant)

        msg = str(exc_info.value)
        for param, annotation in ValidationAssistant.answer.__annotations__.items():
            assert f"{param}: {annotation.__name__}" in msg

    def test_params_validation_wrong_type(self, demo_document):
        class ValidationAssistant(Assistant):
            def answer(
                self,
                messages,
                bool_param: bool,
                int_param: int,
                float_param: float,
                string_param: str,
            ):
                pass

        params = {
            "bool_param": 1,
            "int_param": 0.5,
            "float_param": "arbitrary_value",
            "string_param": False,
        }

        with pytest.raises(RagnaException, match="wrong type") as exc_info:
            self.chat(
                documents=[demo_document], assistant=ValidationAssistant, **params
            )

        msg = str(exc_info.value)
        for param, value in params.items():
            annotation = ValidationAssistant.answer.__annotations__[param]
            assert f"{param}: {annotation.__name__} = {value!r}" in msg

    def test_document_path(self, demo_document):
        chat = self.chat(documents=[demo_document.path])

        assert len(chat.documents) == 1
        document = chat.documents[0]

        assert isinstance(document, LocalDocument)
        assert document.path == demo_document.path
        assert document.name == demo_document.name



from __future__ import annotations

import contextlib
import datetime
import itertools
import uuid
from collections import defaultdict
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    Generic,
    Iterable,
    Iterator,
    Optional,
    TypeVar,
    Union,
    cast,
)

import pydantic
import pydantic_core
from fastapi import status

from ragna._utils import as_async_iterator, as_awaitable, default_user

from ._components import Assistant, Component, Message, MessageRole, SourceStorage
from ._document import Document, LocalDocument
from ._utils import RagnaException, merge_models

if TYPE_CHECKING:
    from ragna.deploy import Config

T = TypeVar("T")
C = TypeVar("C", bound=Component, covariant=True)


class Rag(Generic[C]):
    """RAG workflow.

    !!! tip

        This class can be imported from `ragna` directly, e.g.

        ```python
        from ragna import Rag
        ```
    """

    def __init__(
        self,
        *,
        config: Optional[Config] = None,
        ignore_unavailable_components: bool = False,
    ) -> None:
        self._components: dict[type[C], C] = {}
        self._display_name_map: dict[str, type[C]] = {}

        if config is not None:
            self._preload_components(
                config=config,
                ignore_unavailable_components=ignore_unavailable_components,
            )

    def _preload_components(
        self, *, config: Config, ignore_unavailable_components: bool
    ) -> None:
        for components in [config.source_storages, config.assistants]:
            components = cast(list[type[Component]], components)
            at_least_one = False
            for component in components:
                loaded_component = self._load_component(
                    component,  #  type: ignore[arg-type]
                    ignore_unavailable=ignore_unavailable_components,
                )
                if loaded_component is None:
                    print(
                        f"Ignoring {component.display_name()}, because it is not available."
                    )
                else:
                    at_least_one = True

            if not at_least_one:
                raise RagnaException(
                    "No component available",
                    components=[component.display_name() for component in components],
                )

    def _load_component(
        self, component: Union[C, type[C], str], *, ignore_unavailable: bool = False
    ) -> Optional[C]:
        cls: type[C]
        instance: Optional[C]

        if isinstance(component, Component):
            instance = cast(C, component)
            cls = type(instance)
        elif isinstance(component, type) and issubclass(component, Component):
            cls = component
            instance = None
        elif isinstance(component, str):
            try:
                cls = self._display_name_map[component]
            except KeyError:
                raise RagnaException(
                    "Unknown component",
                    display_name=component,
                    help="Did you forget to create the Rag() instance with a config?",
                    http_status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                    http_detail=f"Unknown component '{component}'",
                ) from None

            instance = None
        else:
            raise RagnaException

        if cls not in self._components:
            if instance is None:
                if not cls.is_available():
                    if ignore_unavailable:
                        return None

                    raise RagnaException(
                        "Component not available", name=cls.display_name()
                    )
                instance = cls()

            self._components[cls] = instance
            self._display_name_map[cls.display_name()] = cls

        return self._components[cls]

    def chat(
        self,
        *,
        documents: Iterable[Any],
        source_storage: Union[SourceStorage, type[SourceStorage], str],
        assistant: Union[Assistant, type[Assistant], str],
        **params: Any,
    ) -> Chat:
        """Create a new [ragna.core.Chat][].

        Args:
            documents: Documents to use. If any item is not a [ragna.core.Document][],
                [ragna.core.LocalDocument.from_path][] is invoked on it.
            source_storage: Source storage to use.
            assistant: Assistant to use.
            **params: Additional parameters passed to the source storage and assistant.
        """
        return Chat(
            self,
            documents=documents,
            source_storage=cast(SourceStorage, self._load_component(source_storage)),  # type: ignore[arg-type]
            assistant=cast(Assistant, self._load_component(assistant)),  # type: ignore[arg-type]
            **params,
        )


class SpecialChatParams(pydantic.BaseModel):
    user: str = pydantic.Field(default_factory=default_user)
    chat_id: uuid.UUID = pydantic.Field(default_factory=uuid.uuid4)
    chat_name: str = pydantic.Field(
        default_factory=lambda: f"Chat {datetime.datetime.now():%x %X}"
    )


class Chat:
    """
    !!! tip

        This object is usually not instantiated manually, but rather through
        [ragna.core.Rag.chat][].

    A chat needs to be [`prepare`][ragna.core.Chat.prepare]d before prompts can be
    [`answer`][ragna.core.Chat.answer]ed.

    Can be used as context manager to automatically invoke
    [`prepare`][ragna.core.Chat.prepare]:

    ```python
    rag = Rag()

    async with rag.chat(
        documents=[path],
        source_storage=ragna.core.RagnaDemoSourceStorage,
        assistant=ragna.core.RagnaDemoAssistant,
    ) as chat:
        print(await chat.answer("What is Ragna?"))
    ```

    Args:
        rag: The RAG workflow this chat is associated with.
        documents: Documents to use. If any item is not a [ragna.core.Document][],
            [ragna.core.LocalDocument.from_path][] is invoked on it.
        source_storage: Source storage to use.
        assistant: Assistant to use.
        **params: Additional parameters passed to the source storage and assistant.
    """

    def __init__(
        self,
        rag: Rag,
        *,
        documents: Iterable[Any],
        source_storage: SourceStorage,
        assistant: Assistant,
        **params: Any,
    ) -> None:
        self._rag = rag

        self.documents = self._parse_documents(documents)
        self.source_storage = source_storage
        self.assistant = assistant

        special_params = SpecialChatParams().model_dump()
        special_params.update(params)
        params = special_params
        self.params = params
        self._unpacked_params = self._unpack_chat_params(params)

        self._prepared = False
        self._messages: list[Message] = []

    async def prepare(self) -> Message:
        """Prepare the chat.

        This [`store`][ragna.core.SourceStorage.store]s the documents in the selected
        source storage. Afterwards prompts can be [`answer`][ragna.core.Chat.answer]ed.

        Returns:
            Welcome message.

        Raises:
            ragna.core.RagnaException: If chat is already
                [`prepare`][ragna.core.Chat.prepare]d.
        """
        if self._prepared:
            raise RagnaException(
                "Chat is already prepared",
                chat=self,
                http_status_code=status.HTTP_400_BAD_REQUEST,
                detail=RagnaException.EVENT,
            )

        await self._as_awaitable(self.source_storage.store, self.documents)
        self._prepared = True

        welcome = Message(
            content="How can I help you with the documents?",
            role=MessageRole.SYSTEM,
        )
        self._messages.append(welcome)
        return welcome

    async def answer(self, prompt: str, *, stream: bool = False) -> Message:
        """Answer a prompt.

        Returns:
            Answer.

        Raises:
            ragna.core.RagnaException: If chat is not
                [`prepare`][ragna.core.Chat.prepare]d.
        """
        if not self._prepared:
            raise RagnaException(
                "Chat is not prepared",
                chat=self,
                http_status_code=status.HTTP_400_BAD_REQUEST,
                detail=RagnaException.EVENT,
            )

        sources = await self._as_awaitable(
            self.source_storage.retrieve, self.documents, prompt
        )

        question = Message(content=prompt, role=MessageRole.USER, sources=sources)
        self._messages.append(question)

        answer = Message(
            content=self._as_async_iterator(
                self.assistant.answer, self._messages.copy()
            ),
            role=MessageRole.ASSISTANT,
            sources=sources,
        )
        if not stream:
            await answer.read()

        self._messages.append(answer)

        return answer

    def _parse_documents(self, documents: Iterable[Any]) -> list[Document]:
        return [
            (
                document
                if isinstance(document, Document)
                else LocalDocument.from_path(document)
            )
            for document in documents
        ]

    def _unpack_chat_params(
        self, params: dict[str, Any]
    ) -> dict[Callable, dict[str, Any]]:
        # This method does two things:
        # 1. Validate the **params against the signatures of the protocol methods of the
        #    used components. This makes sure that
        #    - No parameter is passed that isn't used by at least one component
        #    - No parameter is missing that is needed by at least one component
        #    - No parameter is passed in the wrong type
        # 2. Prepare the distribution of the parameters to the protocol method that
        #    requested them. The actual distribution happens in self._run and
        #    self._run_gen, but is only a dictionary lookup by then.
        component_models = {
            getattr(component, name): model
            for component in [self.source_storage, self.assistant]
            for (_, name), model in component._protocol_models().items()
        }

        ChatModel = merge_models(
            f"{self.__module__}.{type(self).__name__}-{self.params['chat_id']}",
            SpecialChatParams,
            *component_models.values(),
            config=pydantic.ConfigDict(extra="forbid"),
        )

        with self._format_validation_error(ChatModel):
            chat_model = ChatModel.model_validate(params, strict=True)

        chat_params = chat_model.model_dump(exclude_none=True)
        return {
            fn: model(**chat_params).model_dump()
            for fn, model in component_models.items()
        }

    @contextlib.contextmanager
    def _format_validation_error(
        self, model_cls: type[pydantic.BaseModel]
    ) -> Iterator[None]:
        try:
            yield
        except pydantic.ValidationError as validation_error:
            errors = defaultdict(list)
            for error in validation_error.errors():
                errors[error["type"]].append(error)

            parts = [
                f"Validating the Chat parameters resulted in {validation_error.error_count()} errors:",
                "",
            ]

            def format_error(
                error: pydantic_core.ErrorDetails,
                *,
                annotation: bool = False,
                value: bool = False,
            ) -> str:
                param = cast(str, error["loc"][0])

                formatted_error = f"- {param}"
                if annotation:
                    annotation_ = cast(
                        type, model_cls.__pydantic_fields__[param].annotation
                    ).__name__
                    formatted_error += f": {annotation_}"

                if value:
                    value_ = error["input"]
                    formatted_error += (
                        f" = {value_!r}" if annotation else f"={value_!r}"
                    )

                return formatted_error

            if "extra_forbidden" in errors:
                parts.extend(
                    [
                        "The following parameters are unknown:",
                        "",
                        *[
                            format_error(error, value=True)
                            for error in errors["extra_forbidden"]
                        ],
                        "",
                    ]
                )

            if "missing" in errors:
                parts.extend(
                    [
                        "The following parameters are missing:",
                        "",
                        *[
                            format_error(error, annotation=True)
                            for error in errors["missing"]
                        ],
                        "",
                    ]
                )

            type_errors = ["string_type", "int_type", "float_type", "bool_type"]
            if any(type_error in errors for type_error in type_errors):
                parts.extend(
                    [
                        "The following parameters have the wrong type:",
                        "",
                        *[
                            format_error(error, annotation=True, value=True)
                            for error in itertools.chain.from_iterable(
                                errors[type_error] for type_error in type_errors
                            )
                        ],
                        "",
                    ]
                )

            raise RagnaException("\n".join(parts))

    def _as_awaitable(
        self, fn: Union[Callable[..., T], Callable[..., Awaitable[T]]], *args: Any
    ) -> Awaitable[T]:
        return as_awaitable(fn, *args, **self._unpacked_params[fn])

    def _as_async_iterator(
        self,
        fn: Union[Callable[..., Iterator[T]], Callable[..., AsyncIterator[T]]],
        *args: Any,
    ) -> AsyncIterator[T]:
        return as_async_iterator(fn, *args, **self._unpacked_params[fn])

    async def __aenter__(self) -> Chat:
        await self.prepare()
        return self

    async def __aexit__(
        self, exc_type: type[Exception], exc: Exception, traceback: str
    ) -> None:
        pass



import contextlib
from pathlib import Path
from typing import Any

from jinja2 import Environment, FileSystemLoader, TemplateNotFound

ENVIRONMENT = Environment(loader=FileSystemLoader(Path(__file__).parent))


def render(template: str, **context: Any) -> str:
    with contextlib.suppress(TemplateNotFound):
        css_template = ENVIRONMENT.get_template(str(Path(template).with_suffix(".css")))
        context["__template_css__"] = css_template.render(**context)

    template = ENVIRONMENT.get_template(template)
    return template.render(**context)



import pytest
from fastapi import status

from ragna import assistants
from ragna.core import RagnaException
from ragna.deploy import Config
from tests.deploy.utils import make_api_app, make_api_client


@pytest.mark.parametrize("ignore_unavailable_components", [True, False])
def test_ignore_unavailable_components(ignore_unavailable_components):
    available_assistant = assistants.RagnaDemoAssistant
    assert available_assistant.is_available()

    unavailable_assistant = assistants.Gpt4
    assert not unavailable_assistant.is_available()

    config = Config(assistants=[available_assistant, unavailable_assistant])

    if ignore_unavailable_components:
        with make_api_client(
            config=config,
            ignore_unavailable_components=ignore_unavailable_components,
        ) as client:
            components = client.get("/api/components").raise_for_status().json()
            assert [assistant["title"] for assistant in components["assistants"]] == [
                available_assistant.display_name()
            ]
    else:
        with pytest.raises(RagnaException, match="not available"):
            make_api_app(
                config=config,
                ignore_unavailable_components=ignore_unavailable_components,
            )


def test_ignore_unavailable_components_at_least_one():
    unavailable_assistant = assistants.Gpt4
    assert not unavailable_assistant.is_available()

    config = Config(assistants=[unavailable_assistant])

    with pytest.raises(RagnaException, match="No component available"):
        make_api_app(
            config=config,
            ignore_unavailable_components=True,
        )


def test_unknown_component(tmp_local_root):
    config = Config(local_root=tmp_local_root)

    document_root = config.local_root / "documents"
    document_root.mkdir()
    document_path = document_root / "test.txt"
    with open(document_path, "w") as file:
        file.write("!\n")

    with make_api_client(
        config=Config(), ignore_unavailable_components=False
    ) as client:
        document = (
            client.post("/api/documents", json=[{"name": document_path.name}])
            .raise_for_status()
            .json()[0]
        )

        with open(document_path, "rb") as file:
            client.put("/api/documents", files={"documents": (document["id"], file)})

        response = client.post(
            "/api/chats",
            json={
                "name": "test-chat",
                "document_ids": [document["id"]],
                "source_storage": "unknown_source_storage",
                "assistant": "unknown_assistant",
            },
        )

        assert response.status_code == status.HTTP_422_UNPROCESSABLE_ENTITY

        error = response.json()["error"]
        assert "Unknown component" in error["message"]



from typing import AsyncIterator, cast

from ragna.core import Message, PackageRequirement, RagnaException, Requirement, Source

from ._http_api import HttpApiAssistant, HttpStreamingProtocol


class AnthropicAssistant(HttpApiAssistant):
    _API_KEY_ENV_VAR = "ANTHROPIC_API_KEY"
    _STREAMING_PROTOCOL = HttpStreamingProtocol.SSE
    _MODEL: str

    @classmethod
    def _extra_requirements(cls) -> list[Requirement]:
        return [PackageRequirement("httpx_sse")]

    @classmethod
    def display_name(cls) -> str:
        return f"Anthropic/{cls._MODEL}"

    def _instructize_system_prompt(self, sources: list[Source]) -> str:
        # See https://docs.anthropic.com/claude/docs/system-prompts
        # See https://docs.anthropic.com/claude/docs/long-context-window-tips#tips-for-document-qa
        instruction = (
            f"I'm going to give you {len(sources)} document(s). "
            f"Read the document(s) carefully because I'm going to ask you a question about them. "
            f"If you can't answer the question with just the given document(s), just say so. "
            "Don't try to make up an answer.\n\n"
        )
        # See https://docs.anthropic.com/claude/docs/long-context-window-tips#structuring-long-documents

        return (
            instruction
            + "<documents>"
            + "\n".join(f"<document>{source.content}</document>" for source in sources)
            + "</documents>"
        )

    async def answer(
        self, messages: list[Message], *, max_new_tokens: int = 256
    ) -> AsyncIterator[str]:
        # See https://docs.anthropic.com/claude/reference/messages_post
        # See https://docs.anthropic.com/claude/reference/streaming
        prompt, sources = (message := messages[-1]).content, message.sources
        async with self._call_api(
            "POST",
            "https://api.anthropic.com/v1/messages",
            headers={
                "accept": "application/json",
                "anthropic-version": "2023-06-01",
                "content-type": "application/json",
                "x-api-key": self._api_key,
            },
            json={
                "model": self._MODEL,
                "system": self._instructize_system_prompt(sources),
                "messages": [{"role": "user", "content": prompt}],
                "max_tokens": max_new_tokens,
                "temperature": 0.0,
                "stream": True,
            },
        ) as stream:
            async for data in stream:
                # See https://docs.anthropic.com/claude/reference/messages-streaming#raw-http-stream-response
                if "error" in data:
                    raise RagnaException(data["error"].pop("message"), **data["error"])
                elif data["type"] == "message_stop":
                    break
                elif data["type"] != "content_block_delta":
                    continue

                yield cast(str, data["delta"].pop("text"))


class ClaudeOpus(AnthropicAssistant):
    """[Claude 3 Opus](https://docs.anthropic.com/claude/docs/models-overview)

    !!! info "Required environment variables"

        - `ANTHROPIC_API_KEY`

    !!! info "Required packages"

        - `httpx_sse`
    """

    _MODEL = "claude-3-opus-20240229"


class ClaudeSonnet(AnthropicAssistant):
    """[Claude 3 Sonnet](https://docs.anthropic.com/claude/docs/models-overview)

    !!! info "Required environment variables"

        - `ANTHROPIC_API_KEY`

    !!! info "Required packages"

        - `httpx_sse`
    """

    _MODEL = "claude-3-sonnet-20240229"


class ClaudeHaiku(AnthropicAssistant):
    """[Claude 3 Haiku](https://docs.anthropic.com/claude/docs/models-overview)

    !!! info "Required environment variables"

        - `ANTHROPIC_API_KEY`

    !!! info "Required packages"

        - `httpx_sse`
    """

    _MODEL = "claude-3-haiku-20240307"



__all__ = [
    "Auth",
    "Config",
    "DummyBasicAuth",
    "GithubOAuth",
    "InMemoryKeyValueStore",
    "JupyterhubServerProxyAuth",
    "KeyValueStore",
    "NoAuth",
    "RedisKeyValueStore",
]

from ._auth import Auth, DummyBasicAuth, GithubOAuth, JupyterhubServerProxyAuth, NoAuth
from ._config import Config
from ._key_value_store import InMemoryKeyValueStore, KeyValueStore, RedisKeyValueStore

# isort: split

from ragna._utils import fix_module

fix_module(globals())
del fix_module



import datetime
import json

import httpx


def main():
    client = httpx.Client(base_url="http://127.0.0.1:31476")
    client.get("/health").raise_for_status()

    ## authentication

    # This only works if Ragna was deployed with ragna.core.NoAuth
    # If that is not the case, login in whatever way is required, grab the API token and
    # use the following instead
    # client.headers["Authorization"] = f"Bearer {api_token}"

    client.get("/login", follow_redirects=True).raise_for_status()

    ## documents

    documents = (
        client.post(
            "/api/documents", json=[{"name": f"document{i}.txt"} for i in range(5)]
        )
        .raise_for_status()
        .json()
    )

    client.put(
        "/api/documents",
        files=[
            ("documents", (document["id"], f"Content of {document['name']}".encode()))
            for document in documents
        ],
    ).raise_for_status()

    ## chat 1

    chat = (
        client.post(
            "/api/chats",
            json={
                "name": "Test chat",
                "document_ids": [document["id"] for document in documents[:2]],
                "source_storage": "Ragna/DemoSourceStorage",
                "assistant": "Ragna/DemoAssistant",
            },
        )
        .raise_for_status()
        .json()
    )

    client.post(f"/api/chats/{chat['id']}/prepare").raise_for_status()
    client.post(
        f"/api/chats/{chat['id']}/answer",
        json={"prompt": "Hello!"},
    ).raise_for_status()

    ## chat 2

    chat = (
        client.post(
            "/api/chats",
            json={
                "name": f"Chat {datetime.datetime.now():%x %X}",
                "document_ids": [document["id"] for document in documents[2:]],
                "source_storage": "Ragna/DemoSourceStorage",
                "assistant": "Ragna/DemoAssistant",
            },
        )
        .raise_for_status()
        .json()
    )
    client.post(f"/api/chats/{chat['id']}/prepare").raise_for_status()
    for _ in range(3):
        client.post(
            f"/api/chats/{chat['id']}/answer",
            json={"prompt": "What is Ragna? Please, I need to know!"},
        ).raise_for_status()

    # ## chat 3

    chat = (
        client.post(
            "/api/chats",
            json={
                "name": (
                    "Really long chat name that likely needs to be truncated somehow. "
                    "If you can read this, truncating failed :boom:"
                ),
                "document_ids": [documents[i]["id"] for i in [0, 2, 4]],
                "source_storage": "Ragna/DemoSourceStorage",
                "assistant": "Ragna/DemoAssistant",
            },
        )
        .raise_for_status()
        .json()
    )
    client.post(f"/api/chats/{chat['id']}/prepare").raise_for_status()
    client.post(
        f"/api/chats/{chat['id']}/answer",
        json={"prompt": "Hello!"},
    ).raise_for_status()
    client.post(
        f"/api/chats/{chat['id']}/answer",
        json={"prompt": "Ok, in that case show me some pretty markdown!"},
    ).raise_for_status()

    chats = client.get("/api/chats").raise_for_status().json()
    print(json.dumps(chats))


if __name__ == "__main__":
    main()



from __future__ import annotations

import uuid
from typing import Any, Collection, Optional, cast
from urllib.parse import urlsplit

from sqlalchemy import create_engine, select
from sqlalchemy.orm import Session, joinedload, sessionmaker

from ragna.core import RagnaException

from . import _orm as orm
from . import _schemas as schemas


class UnknownUser(Exception):
    def __init__(
        self, name: Optional[str] = None, api_key: Optional[str] = None
    ) -> None:
        self.name = name
        self.api_key = api_key


class Database:
    def __init__(self, url: str) -> None:
        components = urlsplit(url)
        if components.scheme == "sqlite":
            connect_args = dict(check_same_thread=False)
        else:
            connect_args = dict()
        engine = create_engine(url, connect_args=connect_args)
        orm.Base.metadata.create_all(bind=engine)

        self.get_session = sessionmaker(autocommit=False, autoflush=False, bind=engine)

        self._to_orm = SchemaToOrmConverter()
        self._to_schema = OrmToSchemaConverter()

    def _get_orm_user_by_name(self, session: Session, *, name: str) -> orm.User:
        user = cast(
            Optional[orm.User],
            session.execute(
                select(orm.User).where(orm.User.name == name)
            ).scalar_one_or_none(),
        )

        if user is None:
            raise UnknownUser(name)

        return user

    def maybe_add_user(self, session: Session, *, user: schemas.User) -> None:
        try:
            self._get_orm_user_by_name(session, name=user.name)
        except UnknownUser:
            orm_user = orm.User(id=uuid.uuid4(), name=user.name)
            session.add(orm_user)
            session.commit()

    def add_api_key(
        self, session: Session, *, user: str, api_key: schemas.ApiKey
    ) -> None:
        user_id = self._get_orm_user_by_name(session, name=user).id
        orm_api_key = orm.ApiKey(
            id=uuid.uuid4(),
            user_id=user_id,
            name=api_key.name,
            value=api_key.value,
            expires_at=api_key.expires_at,
        )
        session.add(orm_api_key)
        session.commit()

    def get_api_keys(self, session: Session, *, user: str) -> list[schemas.ApiKey]:
        return [
            self._to_schema.api_key(api_key)
            for api_key in session.execute(
                select(orm.ApiKey).where(
                    orm.ApiKey.user_id
                    == self._get_orm_user_by_name(session, name=user).id
                )
            )
            .scalars()
            .all()
        ]

    def delete_api_key(self, session: Session, *, user: str, id: uuid.UUID) -> None:
        orm_api_key = session.execute(
            select(orm.ApiKey).where(
                (orm.ApiKey.id == id)
                & (
                    orm.ApiKey.user_id
                    == self._get_orm_user_by_name(session, name=user).id
                )
            )
        ).scalar_one_or_none()

        if orm_api_key is None:
            raise RagnaException

        session.delete(orm_api_key)  # type: ignore[no-untyped-call]
        session.commit()

    def get_user_by_api_key(
        self, session: Session, api_key_value: str
    ) -> Optional[tuple[schemas.User, schemas.ApiKey]]:
        orm_api_key = session.execute(
            select(orm.ApiKey)  # type: ignore[attr-defined]
            .options(joinedload(orm.ApiKey.user))
            .where(orm.ApiKey.value == api_key_value)
        ).scalar_one_or_none()

        if orm_api_key is None:
            return None

        return (
            self._to_schema.user(orm_api_key.user),
            self._to_schema.api_key(orm_api_key),
        )

    def add_documents(
        self,
        session: Session,
        *,
        user: str,
        documents: list[schemas.Document],
    ) -> None:
        user_id = self._get_orm_user_by_name(session, name=user).id
        session.add_all(
            [self._to_orm.document(document, user_id=user_id) for document in documents]
        )
        session.commit()

    def _get_orm_documents(
        self, session: Session, *, user: str, ids: Collection[uuid.UUID]
    ) -> list[orm.Document]:
        # FIXME also check if the user is allowed to access the documents
        # FIXME: maybe just take the user id to avoid getting it twice in add_chat?
        documents = (
            session.execute(select(orm.Document).where(orm.Document.id.in_(ids)))
            .scalars()
            .all()
        )
        if len(documents) != len(ids):
            raise RagnaException(
                str(set(ids) - {document.id for document in documents})
            )

        return documents  # type: ignore[no-any-return]

    def get_documents(
        self, session: Session, *, user: str, ids: Collection[uuid.UUID]
    ) -> list[schemas.Document]:
        return [
            self._to_schema.document(document)
            for document in self._get_orm_documents(session, user=user, ids=ids)
        ]

    def add_chat(self, session: Session, *, user: str, chat: schemas.Chat) -> None:
        orm_chat = self._to_orm.chat(
            chat, user_id=self._get_orm_user_by_name(session, name=user).id
        )
        # We need to merge and not add here, because the documents are already in the DB
        session.merge(orm_chat)
        session.commit()

    def _select_chat(self, *, eager: bool = False) -> Any:
        selector = select(orm.Chat)
        if eager:
            selector = selector.options(  # type: ignore[attr-defined]
                joinedload(orm.Chat.messages).joinedload(orm.Message.sources),
                joinedload(orm.Chat.documents),
            )
        return selector

    def get_chats(self, session: Session, *, user: str) -> list[schemas.Chat]:
        return [
            self._to_schema.chat(chat)
            for chat in session.execute(
                self._select_chat(eager=True).where(
                    orm.Chat.user_id
                    == self._get_orm_user_by_name(session, name=user).id
                )
            )
            .scalars()
            .unique()
            .all()
        ]

    def _get_orm_chat(
        self, session: Session, *, user: str, id: uuid.UUID, eager: bool = False
    ) -> orm.Chat:
        chat: Optional[orm.Chat] = (
            session.execute(
                self._select_chat(eager=eager).where(
                    (orm.Chat.id == id)
                    & (
                        orm.Chat.user_id
                        == self._get_orm_user_by_name(session, name=user).id
                    )
                )
            )
            .unique()
            .scalar_one_or_none()
        )
        if chat is None:
            raise RagnaException()
        return chat

    def get_chat(self, session: Session, *, user: str, id: uuid.UUID) -> schemas.Chat:
        return self._to_schema.chat(
            (self._get_orm_chat(session, user=user, id=id, eager=True))
        )

    def update_chat(self, session: Session, user: str, chat: schemas.Chat) -> None:
        orm_chat = self._to_orm.chat(
            chat, user_id=self._get_orm_user_by_name(session, name=user).id
        )
        session.merge(orm_chat)
        session.commit()

    def delete_chat(self, session: Session, user: str, id: uuid.UUID) -> None:
        orm_chat = self._get_orm_chat(session, user=user, id=id)
        session.delete(orm_chat)  # type: ignore[no-untyped-call]
        session.commit()


class SchemaToOrmConverter:
    def document(
        self, document: schemas.Document, *, user_id: uuid.UUID
    ) -> orm.Document:
        return orm.Document(
            id=document.id,
            user_id=user_id,
            name=document.name,
            metadata_=document.metadata,
        )

    def source(self, source: schemas.Source) -> orm.Source:
        return orm.Source(
            id=source.id,
            document_id=source.document.id,
            location=source.location,
            content=source.content,
            num_tokens=source.num_tokens,
        )

    def message(self, message: schemas.Message, *, chat_id: uuid.UUID) -> orm.Message:
        return orm.Message(
            id=message.id,
            chat_id=chat_id,
            content=message.content,
            role=message.role,
            sources=[self.source(source) for source in message.sources],
            timestamp=message.timestamp,
        )

    def chat(
        self,
        chat: schemas.Chat,
        *,
        user_id: uuid.UUID,
    ) -> orm.Chat:
        return orm.Chat(
            id=chat.id,
            user_id=user_id,
            name=chat.name,
            documents=[
                self.document(document, user_id=user_id) for document in chat.documents
            ],
            source_storage=chat.source_storage,
            assistant=chat.assistant,
            params=chat.params,
            messages=[
                self.message(message, chat_id=chat.id) for message in chat.messages
            ],
            prepared=chat.prepared,
        )


class OrmToSchemaConverter:
    def user(self, user: orm.User) -> schemas.User:
        return schemas.User(name=user.name)

    def api_key(self, api_key: orm.ApiKey) -> schemas.ApiKey:
        return schemas.ApiKey(
            id=api_key.id,
            name=api_key.name,
            expires_at=api_key.expires_at,
            obfuscated=True,
            value=api_key.value,
        )

    def document(self, document: orm.Document) -> schemas.Document:
        return schemas.Document(
            id=document.id, name=document.name, metadata=document.metadata_
        )

    def source(self, source: orm.Source) -> schemas.Source:
        return schemas.Source(
            id=source.id,
            document=self.document(source.document),
            location=source.location,
            content=source.content,
            num_tokens=source.num_tokens,
        )

    def message(self, message: orm.Message) -> schemas.Message:
        return schemas.Message(
            id=message.id,
            role=message.role,  # type: ignore[arg-type]
            content=message.content,
            sources=[self.source(source) for source in message.sources],
            timestamp=message.timestamp,
        )

    def chat(self, chat: orm.Chat) -> schemas.Chat:
        return schemas.Chat(
            id=chat.id,
            name=chat.name,
            documents=[self.document(document) for document in chat.documents],
            source_storage=chat.source_storage,
            assistant=chat.assistant,
            params=chat.params,
            messages=[self.message(message) for message in chat.messages],
            prepared=chat.prepared,
        )



import uuid
from typing import Annotated, AsyncIterator

import pydantic
from fastapi import APIRouter, Body, UploadFile
from fastapi.responses import StreamingResponse

from . import _schemas as schemas
from ._auth import UserDependency
from ._engine import Engine


def make_router(engine: Engine) -> APIRouter:
    router = APIRouter(tags=["API"])

    @router.post("/documents")
    def register_documents(
        user: UserDependency, document_registrations: list[schemas.DocumentRegistration]
    ) -> list[schemas.Document]:
        return engine.register_documents(
            user=user.name, document_registrations=document_registrations
        )

    @router.put("/documents")
    async def upload_documents(
        user: UserDependency, documents: list[UploadFile]
    ) -> None:
        def make_content_stream(file: UploadFile) -> AsyncIterator[bytes]:
            async def content_stream() -> AsyncIterator[bytes]:
                while content := await file.read(16 * 1024):
                    yield content

            return content_stream()

        await engine.store_documents(
            user=user.name,
            ids_and_streams=[
                (uuid.UUID(document.filename), make_content_stream(document))
                for document in documents
            ],
        )

    @router.get("/components")
    def get_components(_: UserDependency) -> schemas.Components:
        return engine.get_components()

    @router.post("/chats")
    async def create_chat(
        user: UserDependency,
        chat_creation: schemas.ChatCreation,
    ) -> schemas.Chat:
        return engine.create_chat(user=user.name, chat_creation=chat_creation)

    @router.get("/chats")
    async def get_chats(user: UserDependency) -> list[schemas.Chat]:
        return engine.get_chats(user=user.name)

    @router.get("/chats/{id}")
    async def get_chat(user: UserDependency, id: uuid.UUID) -> schemas.Chat:
        return engine.get_chat(user=user.name, id=id)

    @router.post("/chats/{id}/prepare")
    async def prepare_chat(user: UserDependency, id: uuid.UUID) -> schemas.Message:
        return await engine.prepare_chat(user=user.name, id=id)

    @router.post("/chats/{id}/answer")
    async def answer(
        user: UserDependency,
        id: uuid.UUID,
        prompt: Annotated[str, Body(..., embed=True)],
        stream: Annotated[bool, Body(..., embed=True)] = False,
    ) -> schemas.Message:
        message_stream = engine.answer_stream(user=user.name, chat_id=id, prompt=prompt)
        answer = await anext(message_stream)

        if not stream:
            content_chunks = [chunk.content async for chunk in message_stream]
            answer.content += "".join(content_chunks)
            return answer

        async def message_chunks() -> AsyncIterator[schemas.Message]:
            yield answer
            async for chunk in message_stream:
                yield chunk

        async def to_jsonl(
            models: AsyncIterator[pydantic.BaseModel],
        ) -> AsyncIterator[str]:
            async for model in models:
                yield f"{model.model_dump_json()}\n"

        return StreamingResponse(  # type: ignore[return-value]
            to_jsonl(message_chunks())
        )

    @router.delete("/chats/{id}")
    async def delete_chat(user: UserDependency, id: uuid.UUID) -> None:
        engine.delete_chat(user=user.name, id=id)

    return router



import secrets
import uuid
from typing import Any, AsyncIterator, Optional, Type, cast

from fastapi import status as http_status_code

import ragna
from ragna import core
from ragna._utils import make_directory
from ragna.core import Rag, RagnaException
from ragna.core._rag import SpecialChatParams

from . import _schemas as schemas
from ._config import Config
from ._database import Database


class Engine:
    def __init__(self, *, config: Config, ignore_unavailable_components: bool) -> None:
        self._config = config
        ragna.local_root(config.local_root)
        self._documents_root = make_directory(config.local_root / "documents")
        self.supports_store_documents = issubclass(
            self._config.document, ragna.core.LocalDocument
        )

        self._database = Database(url=config.database_url)

        self._rag = Rag(  # type: ignore[var-annotated]
            config=config,
            ignore_unavailable_components=ignore_unavailable_components,
        )

        self._to_core = SchemaToCoreConverter(config=self._config, rag=self._rag)
        self._to_schema = CoreToSchemaConverter()

    def maybe_add_user(self, user: schemas.User) -> None:
        with self._database.get_session() as session:
            return self._database.maybe_add_user(session, user=user)

    def get_user_by_api_key(
        self, api_key_value: str
    ) -> tuple[Optional[schemas.User], bool]:
        with self._database.get_session() as session:
            data = self._database.get_user_by_api_key(
                session, api_key_value=api_key_value
            )

        if data is None:
            return None, False

        user, api_key = data
        return user, api_key.expired

    def create_api_key(
        self, user: str, api_key_creation: schemas.ApiKeyCreation
    ) -> schemas.ApiKey:
        api_key = schemas.ApiKey(
            name=api_key_creation.name,
            expires_at=api_key_creation.expires_at,
            obfuscated=False,
            value=secrets.token_urlsafe(32)[:32],
        )

        with self._database.get_session() as session:
            self._database.add_api_key(session, user=user, api_key=api_key)

        return api_key

    def list_api_keys(self, user: str) -> list[schemas.ApiKey]:
        with self._database.get_session() as session:
            return self._database.get_api_keys(session, user=user)

    def delete_api_key(self, user: str, id: uuid.UUID) -> None:
        with self._database.get_session() as session:
            self._database.delete_api_key(session, user=user, id=id)

    def _get_component_json_schema(
        self,
        component: Type[core.Component],
    ) -> dict[str, dict[str, Any]]:
        json_schema = component._protocol_model().model_json_schema()
        # FIXME: there is likely a better way to exclude certain fields builtin in
        #  pydantic
        for special_param in SpecialChatParams.__pydantic_fields__:
            if (
                "properties" in json_schema
                and special_param in json_schema["properties"]
            ):
                del json_schema["properties"][special_param]
            if "required" in json_schema and special_param in json_schema["required"]:
                json_schema["required"].remove(special_param)
        return json_schema

    def get_components(self) -> schemas.Components:
        return schemas.Components(
            documents=sorted(self._config.document.supported_suffixes()),
            source_storages=[
                self._get_component_json_schema(source_storage)
                for source_storage in self._rag._components.keys()
                if issubclass(source_storage, core.SourceStorage)
            ],
            assistants=[
                self._get_component_json_schema(assistant)
                for assistant in self._rag._components.keys()
                if issubclass(assistant, core.Assistant)
            ],
        )

    def register_documents(
        self, *, user: str, document_registrations: list[schemas.DocumentRegistration]
    ) -> list[schemas.Document]:
        # We create core.Document's first, because they might update the metadata
        core_documents = [
            self._config.document(
                name=registration.name, metadata=registration.metadata
            )
            for registration in document_registrations
        ]
        documents = [self._to_schema.document(document) for document in core_documents]

        with self._database.get_session() as session:
            self._database.add_documents(session, user=user, documents=documents)

        return documents

    async def store_documents(
        self,
        *,
        user: str,
        ids_and_streams: list[tuple[uuid.UUID, AsyncIterator[bytes]]],
    ) -> None:
        if not self.supports_store_documents:
            raise RagnaException(
                "Ragna configuration does not support local upload",
                http_status_code=http_status_code.HTTP_400_BAD_REQUEST,
            )

        ids, streams = zip(*ids_and_streams)

        with self._database.get_session() as session:
            documents = self._database.get_documents(session, user=user, ids=ids)

        for document, stream in zip(documents, streams):
            core_document = cast(
                ragna.core.LocalDocument, self._to_core.document(document)
            )
            await core_document._write(stream)

    def create_chat(
        self, *, user: str, chat_creation: schemas.ChatCreation
    ) -> schemas.Chat:
        params = chat_creation.model_dump()
        document_ids = params.pop("document_ids")
        with self._database.get_session() as session:
            documents = self._database.get_documents(
                session, user=user, ids=document_ids
            )

        chat = schemas.Chat(documents=documents, **params)

        # Although we don't need the actual core.Chat here, this performs the input
        # validation.
        self._to_core.chat(chat, user=user)

        with self._database.get_session() as session:
            self._database.add_chat(session, user=user, chat=chat)

        return chat

    def get_chats(self, *, user: str) -> list[schemas.Chat]:
        with self._database.get_session() as session:
            return self._database.get_chats(session, user=user)

    def get_chat(self, *, user: str, id: uuid.UUID) -> schemas.Chat:
        with self._database.get_session() as session:
            return self._database.get_chat(session, user=user, id=id)

    async def prepare_chat(self, *, user: str, id: uuid.UUID) -> schemas.Message:
        core_chat = self._to_core.chat(self.get_chat(user=user, id=id), user=user)
        core_message = await core_chat.prepare()

        with self._database.get_session() as session:
            self._database.update_chat(
                session, chat=self._to_schema.chat(core_chat), user=user
            )

        return self._to_schema.message(core_message)

    async def answer_stream(
        self, *, user: str, chat_id: uuid.UUID, prompt: str
    ) -> AsyncIterator[schemas.Message]:
        core_chat = self._to_core.chat(self.get_chat(user=user, id=chat_id), user=user)
        core_message = await core_chat.answer(prompt, stream=True)

        content_stream = aiter(core_message)
        content_chunk = await anext(content_stream)
        message = self._to_schema.message(core_message, content_override=content_chunk)
        yield message

        # Avoid sending the sources multiple times
        message_chunk = message.model_copy(update=dict(sources=None))
        async for content_chunk in content_stream:
            message_chunk.content = content_chunk
            yield message_chunk

        with self._database.get_session() as session:
            self._database.update_chat(
                session, chat=self._to_schema.chat(core_chat), user=user
            )

    def delete_chat(self, *, user: str, id: uuid.UUID) -> None:
        with self._database.get_session() as session:
            self._database.delete_chat(session, user=user, id=id)


class SchemaToCoreConverter:
    def __init__(self, *, config: Config, rag: Rag) -> None:
        self._config = config
        self._rag = rag

    def document(self, document: schemas.Document) -> core.Document:
        return self._config.document(
            id=document.id,
            name=document.name,
            metadata=document.metadata,
        )

    def source(self, source: schemas.Source) -> core.Source:
        return core.Source(
            id=source.id,
            document=self.document(source.document),
            location=source.location,
            content=source.content,
            num_tokens=source.num_tokens,
        )

    def message(self, message: schemas.Message) -> core.Message:
        return core.Message(
            message.content,
            role=message.role,
            sources=[self.source(source) for source in message.sources],
        )

    def chat(self, chat: schemas.Chat, *, user: str) -> core.Chat:
        core_chat = self._rag.chat(
            user=user,
            chat_id=chat.id,
            chat_name=chat.name,
            documents=[self.document(document) for document in chat.documents],
            source_storage=chat.source_storage,
            assistant=chat.assistant,
            **chat.params,
        )
        core_chat._messages = [self.message(message) for message in chat.messages]
        core_chat._prepared = chat.prepared

        return core_chat


class CoreToSchemaConverter:
    def document(self, document: core.Document) -> schemas.Document:
        return schemas.Document(
            id=document.id,
            name=document.name,
            metadata=document.metadata,
        )

    def source(self, source: core.Source) -> schemas.Source:
        return schemas.Source(
            id=source.id,
            document=self.document(source.document),
            location=source.location,
            content=source.content,
            num_tokens=source.num_tokens,
        )

    def message(
        self, message: core.Message, *, content_override: Optional[str] = None
    ) -> schemas.Message:
        return schemas.Message(
            id=message.id,
            content=(
                content_override if content_override is not None else message.content
            ),
            role=message.role,
            sources=[self.source(source) for source in message.sources],
            timestamp=message.timestamp,
        )

    def chat(self, chat: core.Chat) -> schemas.Chat:
        params = chat.params.copy()
        del params["user"]
        return schemas.Chat(
            id=params.pop("chat_id"),
            name=params.pop("chat_name"),
            documents=[self.document(document) for document in chat.documents],
            source_storage=chat.source_storage.display_name(),
            assistant=chat.assistant.display_name(),
            params=params,
            messages=[self.message(message) for message in chat._messages],
            prepared=chat._prepared,
        )



"""
# Web UI

Ragna comes with a web UI where you can try out all of the features!

!!! tip "Under development"

    This documentation page is being actively worked on.
    It will be updated with more details soon. :)
"""

# %%
# You can launch the web application from the CLI:
#
# ```bash
# ragna deploy
# ```
#
# The UI will open in a browser window automatically.



import panel as pn
import param

from ragna.deploy._engine import Engine

from . import js
from . import styles as ui
from .api_wrapper import ApiWrapper
from .main_page import MainPage

pn.extension(
    loading_spinner="dots",
    loading_color=ui.MAIN_COLOR,
    layout_compatibility="error",
)
pn.config.browser_info = True


class App(param.Parameterized):
    def __init__(self, engine: Engine):
        super().__init__()

        # Apply the design modifiers to the panel components
        # It returns all the CSS files of the modifiers
        self.css_filepaths = ui.apply_design_modifiers()
        self._engine = engine

    def get_template(self):
        # A bit hacky, but works.
        # we need to preload the css files to avoid a flash of unstyled content, especially when switching between chats.
        # This is achieved by adding <link ref="preload" ...> tags in the head of the document.
        # But none of the panel templates allow to add custom link tags in the head.
        # the only way I found is to take advantage of the raw_css parameter, which allows to add custom css in the head.
        preload_css = "\n".join(
            [
                f"""<link rel="preload" href="{css_fp}" as="style" />"""
                for css_fp in self.css_filepaths
            ]
        )
        preload_css = f"""
                     </style>
                     {preload_css}
                     <style type="text/css">
                     """

        template = pn.template.FastListTemplate(
            # We need to set a title to have it appearing on the browser's tab
            # but it means we need to hide it from the header bar
            title="Ragna",
            accent_base_color=ui.MAIN_COLOR,
            theme_toggle=False,
            collapsed_sidebar=True,
            raw_css=[ui.CSS_VARS, preload_css],
            favicon="imgs/ragna_logo.svg",
            css_files=["https://rsms.me/inter/inter.css", "css/main.css"],
        )

        template.modal.objects = [
            pn.Column(
                sizing_mode="stretch_both",
            )
        ]

        template.header.append(pn.pane.HTML(js.SHADOWROOT_INDEXING))
        template.header.append(pn.pane.HTML(js.MODAL_MOUSE_UP_FIX))
        template.header.append(pn.pane.HTML(js.CONNECTION_MONITOR))

        return template

    def index_page(self):
        api_wrapper = ApiWrapper(self._engine)

        template = self.get_template()
        main_page = MainPage(api_wrapper=api_wrapper, template=template)
        template.main.append(main_page)
        return template

    def health_page(self):
        return pn.pane.HTML("<h1>Ok</h1>")


def app(engine: Engine) -> App:
    return App(engine)



import contextlib
import enum
import json
import os
from typing import Any, AsyncContextManager, AsyncIterator, Optional

import httpx

import ragna
from ragna.core import (
    Assistant,
    EnvVarRequirement,
    PackageRequirement,
    RagnaException,
    Requirement,
)


class HttpStreamingProtocol(enum.Enum):
    SSE = enum.auto()
    JSONL = enum.auto()
    JSON = enum.auto()


class HttpApiCaller:
    @classmethod
    def requirements(cls, protocol: HttpStreamingProtocol) -> list[Requirement]:
        streaming_requirements: dict[HttpStreamingProtocol, list[Requirement]] = {
            HttpStreamingProtocol.SSE: [PackageRequirement("httpx_sse")],
            HttpStreamingProtocol.JSON: [PackageRequirement("ijson")],
        }
        return streaming_requirements.get(protocol, [])

    def __init__(
        self,
        client: httpx.AsyncClient,
        protocol: Optional[HttpStreamingProtocol] = None,
    ) -> None:
        self._client = client
        self._protocol = protocol

    def __call__(
        self,
        method: str,
        url: str,
        *,
        parse_kwargs: Optional[dict[str, Any]] = None,
        **kwargs: Any,
    ) -> AsyncContextManager[AsyncIterator[Any]]:
        if self._protocol is None:
            call_method = self._no_stream
        else:
            call_method = {
                HttpStreamingProtocol.SSE: self._stream_sse,
                HttpStreamingProtocol.JSONL: self._stream_jsonl,
                HttpStreamingProtocol.JSON: self._stream_json,
            }[self._protocol]

        return call_method(method, url, parse_kwargs=parse_kwargs or {}, **kwargs)

    @contextlib.asynccontextmanager
    async def _no_stream(
        self,
        method: str,
        url: str,
        *,
        parse_kwargs: dict[str, Any],
        **kwargs: Any,
    ) -> AsyncIterator[Any]:
        response = await self._client.request(method, url, **kwargs)
        await self._assert_api_call_is_success(response)

        async def stream() -> AsyncIterator[Any]:
            yield response.json()

        yield stream()

    @contextlib.asynccontextmanager
    async def _stream_sse(
        self,
        method: str,
        url: str,
        *,
        parse_kwargs: dict[str, Any],
        **kwargs: Any,
    ) -> AsyncIterator[Any]:
        import httpx_sse

        async with httpx_sse.aconnect_sse(
            self._client, method, url, **kwargs
        ) as event_source:
            await self._assert_api_call_is_success(event_source.response)

            async def stream() -> AsyncIterator[Any]:
                async for sse in event_source.aiter_sse():
                    yield json.loads(sse.data)

            yield stream()

    @contextlib.asynccontextmanager
    async def _stream_jsonl(
        self,
        method: str,
        url: str,
        *,
        parse_kwargs: dict[str, Any],
        **kwargs: Any,
    ) -> AsyncIterator[Any]:
        async with self._client.stream(method, url, **kwargs) as response:
            await self._assert_api_call_is_success(response)

            async def stream() -> AsyncIterator[Any]:
                async for chunk in response.aiter_lines():
                    yield json.loads(chunk)

            yield stream()

    # ijson does not support reading from an (async) iterator, but only from file-like
    # objects, i.e. https://docs.python.org/3/tutorial/inputoutput.html#methods-of-file-objects.
    # See https://github.com/ICRAR/ijson/issues/44 for details.
    # ijson actually doesn't care about most of the file interface and only requires the
    # read() method to be present.
    class _AsyncIteratorReader:
        def __init__(self, ait: AsyncIterator[bytes]) -> None:
            self._ait = ait

        async def read(self, n: int) -> bytes:
            # n is usually used to indicate how many bytes to read, but since we want to
            # return a chunk as soon as it is available, we ignore the value of n. The
            # only exception is n == 0, which is used by ijson to probe the return type
            # and set up decoding.
            if n == 0:
                return b""
            return await anext(self._ait, b"")

    @contextlib.asynccontextmanager
    async def _stream_json(
        self,
        method: str,
        url: str,
        *,
        parse_kwargs: dict[str, Any],
        **kwargs: Any,
    ) -> AsyncIterator[Any]:
        import ijson

        item = parse_kwargs["item"]
        chunk_size = parse_kwargs.get("chunk_size", 16)

        async with self._client.stream(method, url, **kwargs) as response:
            await self._assert_api_call_is_success(response)

            async def stream() -> AsyncIterator[Any]:
                async for chunk in ijson.items(
                    self._AsyncIteratorReader(response.aiter_bytes(chunk_size)), item
                ):
                    yield chunk

            yield stream()

    async def _assert_api_call_is_success(self, response: httpx.Response) -> None:
        if response.is_success:
            return

        content = await response.aread()
        with contextlib.suppress(Exception):
            content = json.loads(content)

        raise RagnaException(
            "API call failed",
            request_method=response.request.method,
            request_url=str(response.request.url),
            response_status_code=response.status_code,
            response_content=content,
        )


class HttpApiAssistant(Assistant):
    _API_KEY_ENV_VAR: Optional[str]
    _STREAMING_PROTOCOL: Optional[HttpStreamingProtocol]

    @classmethod
    def requirements(cls) -> list[Requirement]:
        requirements: list[Requirement] = (
            [EnvVarRequirement(cls._API_KEY_ENV_VAR)]
            if cls._API_KEY_ENV_VAR is not None
            else []
        )
        if cls._STREAMING_PROTOCOL is not None:
            requirements.extend(HttpApiCaller.requirements(cls._STREAMING_PROTOCOL))
        requirements.extend(cls._extra_requirements())
        return requirements

    @classmethod
    def _extra_requirements(cls) -> list[Requirement]:
        return []

    def __init__(self) -> None:
        self._client = httpx.AsyncClient(
            headers={"User-Agent": f"{ragna.__version__}/{self}"},
            timeout=60,
        )
        self._api_key: Optional[str] = (
            os.environ[self._API_KEY_ENV_VAR]
            if self._API_KEY_ENV_VAR is not None
            else None
        )
        self._call_api = HttpApiCaller(self._client, self._STREAMING_PROTOCOL)



import panel as pn
import param


class RightSidebar(pn.viewable.Viewer):
    title = param.String(default="")
    content = param.List(default=[])

    def __init__(self, **params):
        super().__init__(**params)

        self.main_column = None
        self.close_button = None

    def show(self):
        self.main_column.css_classes = ["right_sidebar_main_column", "visible_sidebar"]

    def hide(self, event):
        self.main_column.css_classes = ["right_sidebar_main_column", "hidden_sidebar"]

    @pn.depends("content")
    def content_layout(self):
        return pn.Column(*self.content, css_classes=["right_sidebar_content"])

    @pn.depends("title")
    def header(self):
        return pn.pane.Markdown(
            f"## {self.title}",
            css_classes=["right_sidebar_header"],
        )

    def __panel__(self):
        self.close_button = pn.widgets.Button(
            icon="x",
            button_type="light",
            css_classes=["right_sidebar_close_button"],
        )
        self.close_button.on_click(self.hide)

        self.main_column = pn.Column(
            self.close_button,
            self.header,
            self.content_layout,
            css_classes=["right_sidebar_main_column"],
        )

        return self.main_column



import contextlib
import time

from fastapi.testclient import TestClient

from ragna.assistants import RagnaDemoAssistant
from ragna.deploy._auth import SessionMiddleware
from ragna.deploy._core import make_app


class TestAssistant(RagnaDemoAssistant):
    def answer(self, messages, *, multiple_answer_chunks: bool = True):
        # Simulate a "real" assistant through a small delay. See
        # https://github.com/Quansight/ragna/pull/401#issuecomment-2095851440
        # for why this is needed.
        #
        # Note: multiple_answer_chunks is given a default value here to satisfy
        # the tests in deploy/ui/test_ui.py. This can be removed if TestAssistant
        # is ever removed from that file.
        time.sleep(1e-3)
        content = next(super().answer(messages))

        if multiple_answer_chunks:
            for chunk in content.split(" "):
                yield f"{chunk} "
        else:
            yield content


def make_api_app(*, config, ignore_unavailable_components):
    return make_app(
        config,
        api=True,
        ui=False,
        ignore_unavailable_components=ignore_unavailable_components,
        open_browser=False,
    )


def authenticate_with_api(client: TestClient) -> None:
    client.get("/login", follow_redirects=True).raise_for_status()
    assert SessionMiddleware._COOKIE_NAME in client.cookies


@contextlib.contextmanager
def make_api_client(*, config, ignore_unavailable_components):
    with TestClient(
        make_api_app(
            config=config,
            ignore_unavailable_components=ignore_unavailable_components,
        )
    ) as client:
        authenticate_with_api(client)
        yield client



"""
# REST API

Ragna was designed to help you quickly build custom RAG powered web
applications. For this you can leverage the built-in
[REST API](../../references/deploy.md).

This tutorial walks you through basic steps of using Ragna's REST API.
"""

# %%
# ## Step 1: Start the REST API
#
# Ragnas REST API is normally started from a terminal with
#
# ```bash
# $ ragna deploy
# ```
#
# For this tutorial we use our helper that does the equivalent just from Python.
#
# !!! note
#
#     By default, the REST API is started from the `ragna.toml` configuration file in
#     the current working directory. If you don't have a configuration file yet, you can
#     run
#
#     ```bash
#     $ ragna init
#     ```
#
#     to start an interactive wizard that helps you create one. The config that we'll
#     be using for this tutorial is equivalent of picking the first option the wizard
#     offers you, i.e. using only demo components.

import ragna._docs as ragna_docs

from ragna.deploy import Config

config = Config()

ragna_deploy = ragna_docs.RagnaDeploy(config=config)

# %%
# Let's make sure the REST API is started correctly and can be reached.

import httpx

client = httpx.Client(base_url=f"http://{config.hostname}:{config.port}")
client.get("/health").raise_for_status()


# %%
# ## Step 2: Authentication
#
# In order to use Ragna's REST API, we need to authenticate first. This is handled by
# the [ragna.deploy.Auth][] class, which can be overridden through the config. By
# default, [ragna.deploy.NoAuth][] is used. By hitting the `/login` endpoint, we get a
# session cookie, which is later used to authorize our requests.

client.get("/login", follow_redirects=True)
dict(client.cookies)

# %%
# !!! note
#
#     In a regular deployment, you'll have login through your browser and create an API
#     key in your profile page. The API key is used as
#     [bearer token](https://swagger.io/docs/specification/authentication/bearer-authentication/)
#     and can be set with
#
#     ```python
#     httpx.Client(..., headers={"Authorization": f"Bearer {RAGNA_API_KEY}"})
#     ```

# %%
# ## Step 3: Uploading documents
#
# Before we start with the upload process, let's first have a look what kind of
# documents are supported.

import json

response = client.get("/api/components").raise_for_status()
print(json.dumps(response.json(), indent=2))

# %%
# For simplicity, let's use a demo document with some information about Ragna

from pathlib import Path

print(ragna_docs.SAMPLE_CONTENT)

document_path = Path.cwd() / "ragna.txt"

with open(document_path, "w") as file:
    file.write(ragna_docs.SAMPLE_CONTENT)

# %%
# The upload process in Ragna consists of two parts:
#
# 1. Register the document in Ragna's database. This returns the document ID, which is
#    needed for the upload.

response = client.post(
    "/api/documents", json=[{"name": document_path.name}]
).raise_for_status()
documents = response.json()
print(json.dumps(documents, indent=2))

# %%
# 2. Perform the upload through a
# [multipart request](https://swagger.io/docs/specification/describing-request-body/multipart-requests/)
# with the following parameters:
#
# - The field is `documents` for all entries
# - The field name is the ID of the document returned by step 1.
# - The field value is the binary content of the document.

client.put(
    "/api/documents",
    files=[("documents", (documents[0]["id"], open(document_path, "rb")))],
)

# %%
# ## Step 4: Select a source storage and assistant
#
# The configuration we are using only supports demo components for the source storage
# and assistant and so we pick them here.

from ragna import source_storages, assistants

source_storage = source_storages.RagnaDemoSourceStorage.display_name()
assistant = assistants.RagnaDemoAssistant.display_name()

print(f"{source_storage=}, {assistant=}")

# %%
# ## Step 5: Start chatting
#
# Now that we have uploaded a document, and selected a source storage and assistant to
# be used, we can create a new chat.

response = client.post(
    "/api/chats",
    json={
        "name": "Tutorial REST API",
        "document_ids": [document["id"] for document in documents],
        "source_storage": source_storage,
        "assistant": assistant,
    },
).raise_for_status()
chat = response.json()
print(json.dumps(chat, indent=2))

# %%
# As can be seen by the `"prepared": false` value in the `chat` JSON object we still
# need to prepare it.

client.post(f"/api/chats/{chat['id']}/prepare").raise_for_status()

# %%
# Finally, we can get answers to our questions.

response = client.post(
    f"/api/chats/{chat['id']}/answer",
    json={"prompt": "What is Ragna?"},
).raise_for_status()
answer = response.json()
print(json.dumps(answer, indent=2))

# %%

print(answer["content"])

# %%
# Before we close the tutorial, let's terminate the REST API and have a look at what
# would have printed in the terminal if we had started it with the `ragna deploy`
# command.

ragna_deploy.terminate()



import asyncio

import panel as pn
import param

from .central_view import CentralView
from .left_sidebar import LeftSidebar
from .modal_configuration import ModalConfiguration
from .right_sidebar import RightSidebar


class MainPage(pn.viewable.Viewer, param.Parameterized):
    current_chat_id = param.String(default=None)
    chats = param.List(default=None)

    def __init__(self, api_wrapper, template):
        super().__init__()
        self.api_wrapper = api_wrapper
        self.template = template

        self.modal = None
        self.central_view = CentralView(api_wrapper=self.api_wrapper)
        self.central_view.on_click_chat_info = (
            lambda event, title, content: self.show_right_sidebar(title, content)
        )

        self.left_sidebar = LeftSidebar(api_wrapper=self.api_wrapper)
        self.left_sidebar.on_click_chat = self.on_click_chat
        self.left_sidebar.on_click_new_chat = lambda event: self.open_modal()

        self.right_sidebar = RightSidebar()

        pn.state.location.sync(
            self,
            {"current_chat_id": "current_chat_id"},
            on_error=lambda x: print(f"error sync on {x}"),
        )

    async def refresh_data(self):
        self.chats = await self.api_wrapper.get_chats()

    @param.depends("chats", watch=True)
    def after_update_chats(self):
        self.left_sidebar.chats = self.chats

        if len(self.chats) > 0:
            chat_id_exist = (
                len([c["id"] for c in self.chats if c["id"] == self.current_chat_id])
                > 0
            )

            if self.current_chat_id is None or not chat_id_exist:
                self.current_chat_id = self.chats[0]["id"]

            for c in self.chats:
                if c["id"] == self.current_chat_id:
                    self.central_view.set_current_chat(c)
                    break

    # Modal and callbacks
    def open_modal(self):
        self.modal = ModalConfiguration(
            api_wrapper=self.api_wrapper,
            new_chat_ready_callback=self.open_new_chat,
            cancel_button_callback=self.on_click_cancel_button,
        )

        self.template.modal.objects[0].objects = [self.modal]
        self.template.open_modal()

    async def open_new_chat(self, new_chat_id):
        # called after creating a new chat.
        self.current_chat_id = new_chat_id
        await self.refresh_data()

        self.template.close_modal()

    def on_click_cancel_button(self, event):
        self.template.close_modal()

    # Left sidebar callbacks
    def on_click_chat(self, chat):
        self.central_view.set_loading(True)
        self.current_chat_id = chat["id"]
        self.central_view.set_current_chat(chat)
        self.central_view.set_loading(False)

    # Right sidebar callbacks
    def show_right_sidebar(self, title, content):
        self.right_sidebar.title = title
        self.right_sidebar.content = content
        self.right_sidebar.param.trigger("content")
        self.right_sidebar.show()

    @param.depends("current_chat_id", watch=True)
    def update_subviews_current_chat_id(self, avoid_senders=[]):
        if self.left_sidebar is not None and self.left_sidebar not in avoid_senders:
            self.left_sidebar.current_chat_id = self.current_chat_id

    def __panel__(self):
        asyncio.ensure_future(self.refresh_data())

        return pn.Row(
            self.left_sidebar,
            self.central_view,
            self.right_sidebar,
            css_classes=["main_page_main_row"],
        )



from typing import AsyncIterator

from ragna.core import Message, Source

from ._http_api import HttpApiAssistant, HttpStreamingProtocol


class GoogleAssistant(HttpApiAssistant):
    _API_KEY_ENV_VAR = "GOOGLE_API_KEY"
    _STREAMING_PROTOCOL = HttpStreamingProtocol.JSON
    _MODEL: str

    @classmethod
    def display_name(cls) -> str:
        return f"Google/{cls._MODEL}"

    def _instructize_prompt(self, prompt: str, sources: list[Source]) -> str:
        # https://ai.google.dev/docs/prompt_best_practices#add-contextual-information
        return "\n".join(
            [
                "Answer the prompt using only the pieces of context below.",
                "If you don't know the answer, just say so. Don't try to make up additional context.",
                f"Prompt: {prompt}",
                *[f"\n{source.content}" for source in sources],
            ]
        )

    async def answer(
        self, messages: list[Message], *, max_new_tokens: int = 256
    ) -> AsyncIterator[str]:
        prompt, sources = (message := messages[-1]).content, message.sources
        async with self._call_api(
            "POST",
            f"https://generativelanguage.googleapis.com/v1beta/models/{self._MODEL}:streamGenerateContent",
            params={"key": self._api_key},
            headers={"Content-Type": "application/json"},
            json={
                "contents": [
                    {"parts": [{"text": self._instructize_prompt(prompt, sources)}]}
                ],
                # https://ai.google.dev/docs/safety_setting_gemini
                "safetySettings": [
                    {
                        "category": f"HARM_CATEGORY_{category}",
                        "threshold": "BLOCK_NONE",
                    }
                    for category in [
                        "HARASSMENT",
                        "HATE_SPEECH",
                        "SEXUALLY_EXPLICIT",
                        "DANGEROUS_CONTENT",
                    ]
                ],
                # https://ai.google.dev/tutorials/rest_quickstart#configuration
                "generationConfig": {
                    "temperature": 0.0,
                    "maxOutputTokens": max_new_tokens,
                },
            },
            parse_kwargs=dict(item="item.candidates.item.content.parts.item.text"),
        ) as stream:
            async for chunk in stream:
                yield chunk


class GeminiPro(GoogleAssistant):
    """[Google Gemini Pro](https://ai.google.dev/models/gemini)

    !!! info "Required environment variables"

        - `GOOGLE_API_KEY`

    !!! info "Required packages"

        - `ijson`
    """

    _MODEL = "gemini-pro"


class GeminiUltra(GoogleAssistant):
    """[Google Gemini Ultra](https://ai.google.dev/models/gemini)

    !!! info "Required environment variables"

        - `GOOGLE_API_KEY`

    !!! info "Required packages"

        - `ijson`
    """

    _MODEL = "gemini-ultra"



import contextlib
import json

import httpx
import pytest


@pytest.fixture(scope="package", autouse=True)
def enhance_raise_for_status(package_mocker):
    raise_for_status = httpx.Response.raise_for_status

    def enhanced_raise_for_status(self):
        __tracebackhide__ = True

        try:
            return raise_for_status(self)
        except httpx.HTTPStatusError as error:
            content = None
            with contextlib.suppress(Exception):
                content = error.response.read()
                content = content.decode()
                content = "\n" + json.dumps(json.loads(content), indent=2)

            if content is None:
                raise error

            message = f"{error}\nResponse content: {content}"
            raise httpx.HTTPStatusError(
                message, request=error.request, response=error.response
            ) from None

    yield package_mocker.patch(
        ".".join(
            [
                httpx.Response.__module__,
                httpx.Response.__name__,
                raise_for_status.__name__,
            ]
        ),
        new=enhanced_raise_for_status,
    )



"""
# Custom Components

While Ragna has builtin support for a few [source storages][ragna.source_storages]
and [assistants][ragna.assistants], its real strength lies in allowing users
to incorporate custom components. This tutorial covers how to do that.
"""

# %%
# ## Components
#
# ### Source Storage
#
# [ragna.core.SourceStorage][]s are objects that take a number of documents and
# [ragna.core.SourceStorage.store][] their content in way such that relevant parts for a
# given user prompt can be [ragna.core.SourceStorage.retrieve][]d in the form of
# [ragna.core.Source][]s. Usually, source storages are vector databases.
#
# In this tutorial, we define a minimal `TutorialSourceStorage` that is similar to
# [ragna.source_storages.RagnaDemoSourceStorage][]. In `.store()` we create the
# `Source`s with the first 100 characters of each document and store them in memory
# based on the unique `chat_id`. In `retrieve()` we return all the stored sources
# for the chat and regardless of the user `prompt`.
#
# !!! note
#
#     The `chat_id` used in both methods will not be passed by default, but rather has
#     to be requested explicitly. How this works in detail will be explained later in
#     this tutorial in the [Custom Parameters](#custom-parameters) section.

import uuid

from ragna.core import Document, Source, SourceStorage, Message


class TutorialSourceStorage(SourceStorage):
    def __init__(self):
        self._storage: dict[uuid.UUID, list[Source]] = {}

    def store(self, documents: list[Document], *, chat_id: uuid.UUID) -> None:
        print(f"Running {type(self).__name__}().store()")

        self._storage[chat_id] = [
            Source(
                id=str(uuid.uuid4()),
                document=document,
                location="N/A",
                content=(content := next(document.extract_pages()).text[:100]),
                num_tokens=len(content.split()),
            )
            for document in documents
        ]

    def retrieve(
        self, documents: list[Document], prompt: str, *, chat_id: uuid.UUID
    ) -> list[Source]:
        print(f"Running {type(self).__name__}().retrieve()")
        return self._storage[chat_id]


# %%
# ### Assistant
#
# [ragna.core.Assistant][]s are objects that take the chat history as list of
# [ragna.core.Message][]s and their relevant [ragna.core.Source][]s and generate a
# response from that. Usually, assistants are LLMs.
#
# In this tutorial, we define a minimal `TutorialAssistant` that is similar to
# [ragna.assistants.RagnaDemoAssistant][]. In `.answer()` we mirror back the user
# prompt and also the number of sources we were given.
#
# !!! note
#
#     The answer needs to be `yield`ed instead of `return`ed. By yielding multiple
#     times, the answer can be streamed back. See the
#     [streaming example](../../generated/examples/gallery_streaming.md) for more
#     information.

from typing import Iterator

from ragna.core import Assistant, Source


class TutorialAssistant(Assistant):
    def answer(self, messages: list[Message]) -> Iterator[str]:
        print(f"Running {type(self).__name__}().answer()")
        # For simplicity, we only deal with the last message here, i.e. the latest user
        # prompt.
        prompt, sources = (message := messages[-1]).content, message.sources
        yield (
            f"To answer the user prompt '{prompt}', "
            f"I was given {len(sources)} source(s)."
        )


# %%
# ## Usage
#
# Now that we have defined a custom [ragna.core.SourceStorage][] and
# [ragna.core.Assistant][], let's have a look on how to use them with Ragna. Let's start
# with the Python API.
#
# ### Python API
#
# We first create a sample document.

from pathlib import Path

import ragna._docs as ragna_docs

print(ragna_docs.SAMPLE_CONTENT)

document_path = Path.cwd() / "ragna.txt"

with open(document_path, "w") as file:
    file.write(ragna_docs.SAMPLE_CONTENT)

# %%
# Next, we create a new [ragna.core.Chat][] with our custom components.

from ragna import Rag

chat = Rag().chat(
    documents=[document_path],
    source_storage=TutorialSourceStorage,
    assistant=TutorialAssistant,
)

# %%
# From here on, you can use your custom components exactly like the builtin ones. For
# more details, have a look at the
# [Python API tutorial](../../generated/tutorials/gallery_python_api.md).

_ = await chat.prepare()

# %%

answer = await chat.answer("What is Ragna?")

# %%

print(answer)

# %%
for idx, source in enumerate(answer.sources, 1):
    print(f"{idx}. {source.content}")

# %%
# ### REST API
#
# To use custom components in Ragna's REST API, we need to include the components in
# the corresponding arrays in the configuration file. If you don't have a `ragna.toml`
# configuration file yet, see the [config reference](../../references/config.md) on how
# to create one.
#
# For example, if we put the `TutorialSourceStorage` and `TutorialAssistant` classes in
# a `tutorial.py` module, we can expand the `source_storages` and `assistants` arrays
# like this
#
# ```toml
# source_storages = [
#     ...
#     "tutorial.TutorialSourceStorage"
# ]
# assistants = [
#     ...
#     "tutorial.TutorialAssistant"
# ]
# ```
#
# !!! note
#
#     Make sure the `tutorial.py` module is on Python's search path. See the
#     [config reference](../../references/config.md#referencing-python-objects) for
#     details.
#
# With the configuration set up, we now start the REST API. For the purpose of this
# tutorial we replicate the same programmatically. For general information on how to use
# the REST API, have a look at the
# [REST API tutorial](../../generated/tutorials/gallery_rest_api.md).

from ragna.deploy import Config

config = Config(
    source_storages=[TutorialSourceStorage],
    assistants=[TutorialAssistant],
)

ragna_deploy = ragna_docs.RagnaDeploy(config)

client, document = ragna_deploy.get_http_client(
    authenticate=True, upload_sample_document=True
)

# %%
# To select our custom components, we pass their display names to the chat creation.
#
# !!! tip
#
#     By default [ragna.core.Component.display_name][] returns the name of the class,
#     but can be overridden, e.g. to format the name better.

import json

response = client.post(
    "/api/chats",
    json={
        "name": "Tutorial Custom Components",
        "document_ids": [document["id"]],
        "source_storage": TutorialSourceStorage.display_name(),
        "assistant": TutorialAssistant.display_name(),
        "params": {},
    },
).raise_for_status()
chat = response.json()

client.post(f"/api/chats/{chat['id']}/prepare").raise_for_status()

response = client.post(
    f"/api/chats/{chat['id']}/answer",
    json={"prompt": "What is Ragna?"},
).raise_for_status()
answer = response.json()
print(json.dumps(answer, indent=2))

# %%
# Let's stop the REST API and have a look at what would have printed in the terminal if
# we had started it with the `ragna api` command.

ragna_deploy.terminate()

# %%
# ### Web UI
#
# The setup for the web UI is exactly the same as for the [REST API](#rest-api). If you
# have your configuration set up, you can start the web UI and use the custom
# components. See the [web UI tutorial](../../generated/tutorials/gallery_rest_api.md)
# for details.
#
#
# ![Create-chat-modal of Ragna's web UI with the TutorialSourceStorage and TutorialAssistant as selected options](../../assets/images/ragna-tutorial-components.png)
#
# !!! warning
#
#     See the section about [custom parameters for the web UI](#web-ui_1) for some
#     limitations using custom components in the web UI.

# %%
# ## Custom parameters
#
# Ragna supports passing parameters to the components on a per-chat level. The extra
# parameters are defined by adding them to the signature of the method. For example,
# let's define a more elaborate assistant that accepts the following arguments:
#
# - `my_required_parameter` must be passed and has to be an `int`
# - `my_optional_parameter` may be passed and has to be a `str` if passed


class ElaborateTutorialAssistant(Assistant):
    def answer(
        self,
        messages: list[Message],
        *,
        my_required_parameter: int,
        my_optional_parameter: str = "foo",
    ) -> Iterator[str]:
        print(f"Running {type(self).__name__}().answer()")
        yield f"I was given {my_required_parameter=} and {my_optional_parameter=}."


# %%
# ### Python API
#
# To pass custom parameters to the components, pass them as keyword arguments when
# creating a chat.

chat = Rag().chat(
    documents=[document_path],
    source_storage=TutorialSourceStorage,
    assistant=ElaborateTutorialAssistant,
    my_required_parameter=3,
    my_optional_parameter="bar",
)

_ = await chat.prepare()
print(await chat.answer("Hello!"))

# %%
# The chat creation will fail if a required parameter is not passed or a wrong type is
# passed for any parameter.

try:
    Rag().chat(
        documents=[document_path],
        source_storage=TutorialSourceStorage,
        assistant=ElaborateTutorialAssistant,
    )
except Exception as exc:
    print(exc)

# %%

try:
    Rag().chat(
        documents=[document_path],
        source_storage=TutorialSourceStorage,
        assistant=ElaborateTutorialAssistant,
        my_required_parameter="bar",
        my_optional_parameter=3,
    )
except Exception as exc:
    print(exc)

# %%
# ### REST API

config = Config(
    source_storages=[TutorialSourceStorage],
    assistants=[ElaborateTutorialAssistant],
)

ragna_deploy = ragna_docs.RagnaDeploy(config)

client, document = ragna_deploy.get_http_client(
    authenticate=True, upload_sample_document=True
)

# %%
# To pass custom parameters, define them in the `params` mapping when creating a new
# chat.

response = client.post(
    "/api/chats",
    json={
        "name": "Tutorial Elaborate Custom Components",
        "document_ids": [document["id"]],
        "source_storage": TutorialSourceStorage.display_name(),
        "assistant": ElaborateTutorialAssistant.display_name(),
        "params": {
            "my_required_parameter": 3,
            "my_optional_parameter": "bar",
        },
    },
).raise_for_status()
chat = response.json()
print(json.dumps(chat, indent=2))

# %%

client.post(f"/api/chats/{chat['id']}/prepare").raise_for_status()

response = client.post(
    f"/api/chats/{chat['id']}/answer",
    json={"prompt": "What is Ragna?"},
).raise_for_status()
answer = response.json()
print(json.dumps(answer, indent=2))

# %%
# Let's stop the REST API and have a look at what would have printed in the terminal if
# we had started it with the `ragna api` command.

ragna_deploy.terminate()

# %%
# ### Web UI
#
# !!! warning
#     Unfortunately, Ragna's web UI currently does **not** support arbitrary custom
#     parameters. This is actively worked on and progress is tracked in
#     [#217](https://github.com/Quansight/ragna/issues/217). Until this is resolved,
#     the source storage and the assistant together take the following four
#     parameters:
#
#     1. `chunk_size: int`
#     2. `chunk_overlap: int`
#     3. `num_tokens: int`
#     4. `max_new_tokens: int`
#
#     When using Ragna's builtin components, all source storages cover parameters 1. to
#     3. and all assistant cover parameter 4.

# %%
# ## Sync or `async`?
#
# So far we have `def`ined all of the abstract methods of the components as regular,
# i.e. synchronous methods. However, this is *not* a requirement. If you want to run
# `async` code inside the methods, define them with `async def` and Ragna will
# handle the rest for you. Internally, Ragna runs synchronous methods on a thread to
# avoid blocking the main thread. Asynchronous methods are scheduled on the main event
# loop.
#
# Let's have a look at a minimal example.

import asyncio
import time
from typing import AsyncIterator


class AsyncAssistant(Assistant):
    async def answer(self, messages: list[Message]) -> AsyncIterator[str]:
        print(f"Running {type(self).__name__}().answer()")
        start = time.perf_counter()
        await asyncio.sleep(0.3)
        stop = time.perf_counter()
        yield f"I've waited for {stop - start} seconds!"


chat = Rag().chat(
    documents=[document_path],
    source_storage=TutorialSourceStorage,
    assistant=AsyncAssistant,
)

_ = await chat.prepare()
print(await chat.answer("Hello!"))



import contextlib
import multiprocessing
import time

import httpx
import pytest
from playwright.sync_api import expect

from ragna._cli.core import deploy as _deploy
from ragna.deploy import Config
from tests.deploy.utils import TestAssistant
from tests.utils import get_available_port


@contextlib.contextmanager
def deploy(config):
    process = multiprocessing.Process(
        target=_deploy,
        kwargs=dict(
            config=config,
            api=False,
            ui=True,
            ignore_unavailable_components=False,
            open_browser=False,
        ),
    )
    try:
        process.start()

        client = httpx.Client(base_url=config._url)

        # FIXME: create a generic utility for this
        def server_available() -> bool:
            try:
                return client.get("/health").is_success
            except httpx.ConnectError:
                return False

        while not server_available():
            time.sleep(0.1)

        yield process
    finally:
        process.terminate()
        process.join()
        process.close()


@pytest.fixture
def default_config(tmp_local_root):
    return Config(
        local_root=tmp_local_root,
        assistants=[TestAssistant],
        port=get_available_port(),
    )


@pytest.fixture
def index_page(default_config, page):
    config = default_config
    with deploy(default_config):
        page.goto(f"http://{config.hostname}:{config.port}/ui")
        yield page


def test_start_chat(index_page, tmp_path) -> None:
    # expect(page.get_by_role("button", name="Sign In")).to_be_visible()

    # # Authorize with no credentials
    # page.get_by_role("button", name="Sign In").click()
    # expect(page.get_by_role("button", name=" New Chat")).to_be_visible()
    #
    # # expect auth token to be set
    # cookies = page.context.cookies()
    # assert len(cookies) == 1
    # cookie = cookies[0]
    # assert cookie.get("name") == "auth_token"
    # auth_token = cookie.get("value")
    # assert auth_token is not None

    # New page button
    new_chat_button = index_page.get_by_role("button", name=" New Chat")
    expect(new_chat_button).to_be_visible()
    new_chat_button.click()

    # document_name = "test.txt"
    # document_path = tmp_path / document_name
    # with open(document_path, "w") as file:
    #     file.write("!\n")

    # # File upload selector
    # with index_page.expect_file_chooser() as fc_info:
    #     index_page.locator(".fileUpload").click()
    # file_chooser = fc_info.value
    # file_chooser.set_files(document_path)

    # # Upload document and expect to see it listed
    # file_list = page.locator(".fileListContainer")
    # expect(file_list.first).to_have_text(str(document_name))
    #
    # chat_dialog = page.get_by_role("dialog")
    # expect(chat_dialog).to_be_visible()
    # start_chat_button = page.get_by_role("button", name="Start Conversation")
    # expect(start_chat_button).to_be_visible()
    # time.sleep(0.5)  # hack while waiting for button to be fully clickable
    # start_chat_button.click(delay=5)
    #
    # chat_box_row = page.locator(".chat-interface-input-row")
    # expect(chat_box_row).to_be_visible()
    #
    # chat_box = chat_box_row.get_by_role("textbox")
    # expect(chat_box).to_be_visible()
    #
    # # Document should be in the database
    # chats_url = f"http://{config.api.hostname}:{config.api.port}/chats"
    # chats = httpx.get(
    #     chats_url, headers={"Authorization": f"Bearer {auth_token}"}
    # ).json()
    # assert len(chats) == 1
    # chat = chats[0]
    # chat_documents = chat["metadata"]["documents"]
    # assert len(chat_documents) == 1
    # assert chat_documents[0]["name"] == document_name
    #
    # chat_box.fill("Tell me about the documents")
    #
    # chat_button = chat_box_row.get_by_role("button")
    # expect(chat_button).to_be_visible()
    # chat_button.click()



import os
from functools import cached_property
from typing import AsyncIterator, cast

from ragna.core import Message, RagnaException

from ._http_api import HttpStreamingProtocol
from ._openai import OpenaiLikeHttpApiAssistant


class OllamaAssistant(OpenaiLikeHttpApiAssistant):
    """[Ollama](https://ollama.com/)

    To use this assistant, start the Ollama server manually. By default, the server
    is expected at `http://localhost:11434`. This can be changed with the
    `RAGNA_OLLAMA_BASE_URL` environment variable.
    """

    _API_KEY_ENV_VAR = None
    _STREAMING_PROTOCOL = HttpStreamingProtocol.JSONL
    _MODEL: str

    @classmethod
    def display_name(cls) -> str:
        return f"Ollama/{cls._MODEL}"

    @cached_property
    def _url(self) -> str:
        base_url = os.environ.get("RAGNA_OLLAMA_BASE_URL", "http://localhost:11434")
        return f"{base_url}/api/chat"

    async def answer(
        self, messages: list[Message], *, max_new_tokens: int = 256
    ) -> AsyncIterator[str]:
        prompt, sources = (message := messages[-1]).content, message.sources
        async with self._call_openai_api(
            prompt, sources, max_new_tokens=max_new_tokens
        ) as stream:
            async for data in stream:
                # Modeled after
                # https://github.com/ollama/ollama/blob/06a1508bfe456e82ba053ea554264e140c5057b5/examples/python-loganalysis/readme.md?plain=1#L57-L62
                if "error" in data:
                    raise RagnaException(data["error"])
                if not data["done"]:
                    yield cast(str, data["message"]["content"])


class OllamaGemma2B(OllamaAssistant):
    """[Gemma:2B](https://ollama.com/library/gemma)"""

    _MODEL = "gemma:2b"


class OllamaLlama2(OllamaAssistant):
    """[Llama 2](https://ollama.com/library/llama2)"""

    _MODEL = "llama2"


class OllamaLlava(OllamaAssistant):
    """[Llava](https://ollama.com/library/llava)"""

    _MODEL = "llava"


class OllamaMistral(OllamaAssistant):
    """[Mistral](https://ollama.com/library/mistral)"""

    _MODEL = "mistral"


class OllamaMixtral(OllamaAssistant):
    """[Mixtral](https://ollama.com/library/mixtral)"""

    _MODEL = "mixtral"


class OllamaOrcaMini(OllamaAssistant):
    """[Orca Mini](https://ollama.com/library/orca-mini)"""

    _MODEL = "orca-mini"


class OllamaPhi2(OllamaAssistant):
    """[Phi-2](https://ollama.com/library/phi)"""

    _MODEL = "phi"



from pathlib import Path
from typing import Annotated, Optional

import rich
import typer
import uvicorn

import ragna
from ragna.deploy._core import make_app

from .config import ConfigOption, check_config, init_config

app = typer.Typer(
    name="Ragna",
    invoke_without_command=True,
    no_args_is_help=True,
    add_completion=False,
    pretty_exceptions_enable=False,
)


def version_callback(value: bool) -> None:
    if value:
        rich.print(f"ragna {ragna.__version__} from {ragna.__path__[0]}")
        raise typer.Exit()


@app.callback()
def _main(
    version: Annotated[
        Optional[bool],
        typer.Option(
            "--version", callback=version_callback, help="Show version and exit."
        ),
    ] = None,
) -> None:
    pass


@app.command(help="Start a wizard to build a Ragna configuration interactively.")
def init(
    *,
    output_path: Annotated[
        Path,
        typer.Option(
            "-o",
            "--output-file",
            metavar="OUTPUT_PATH",
            default_factory=lambda: Path.cwd() / "ragna.toml",
            show_default="./ragna.toml",
            help="Write configuration to <OUTPUT_PATH>.",
        ),
    ],
    force: Annotated[
        bool,
        typer.Option(
            "-f", "--force", help="Overwrite an existing file at <OUTPUT_PATH>."
        ),
    ] = False,
) -> None:
    config, output_path, force = init_config(output_path=output_path, force=force)
    config.to_file(output_path, force=force)


@app.command(help="Check the availability of components.")
def check(config: ConfigOption = "./ragna.toml") -> None:  # type: ignore[assignment]
    is_available = check_config(config)
    raise typer.Exit(int(not is_available))


@app.command(help="Deploy Ragna REST API and web UI.")
def deploy(
    *,
    config: ConfigOption = "./ragna.toml",  # type: ignore[assignment]
    api: Annotated[
        bool,
        typer.Option(
            "--api/--no-api",
            help="Deploy the Ragna REST API.",
        ),
    ] = True,
    ui: Annotated[
        bool,
        typer.Option(
            help="Deploy the Ragna web UI.",
        ),
    ] = True,
    ignore_unavailable_components: Annotated[
        bool,
        typer.Option(
            help=(
                "Ignore components that are not available, "
                "i.e. their requirements are not met. "
            )
        ),
    ] = False,
    open_browser: Annotated[
        Optional[bool],
        typer.Option(
            help="Open a browser when Ragna is deployed.",
            show_default="value of ui / no-ui",
        ),
    ] = None,
) -> None:
    if not (api or ui):
        raise Exception

    if open_browser is None:
        open_browser = ui

    uvicorn.run(
        lambda: make_app(
            config,
            ui=ui,
            api=api,
            ignore_unavailable_components=ignore_unavailable_components,
            open_browser=open_browser,
        ),
        factory=True,
        host=config.hostname,
        port=config.port,
    )



from __future__ import annotations

import abc
import os
import time
from typing import Any, Callable, Generic, Optional, TypeVar, Union, cast

import pydantic

from ragna.core import PackageRequirement, Requirement
from ragna.core._utils import RequirementsMixin

M = TypeVar("M", bound=pydantic.BaseModel)


class SerializableModel(pydantic.BaseModel, Generic[M]):
    cls: pydantic.ImportString[type[M]]
    obj: dict[str, Any]

    @classmethod
    def from_model(cls, model: M) -> SerializableModel[M]:
        return SerializableModel(cls=type(model), obj=model.model_dump(mode="json"))

    def to_model(self) -> M:
        return self.cls.model_validate(self.obj)


class KeyValueStore(abc.ABC, RequirementsMixin, Generic[M]):
    def serialize(self, model: M) -> str:
        return SerializableModel.from_model(model).model_dump_json()

    def deserialize(self, data: Union[str, bytes]) -> M:
        return SerializableModel.model_validate_json(data).to_model()

    @abc.abstractmethod
    def set(
        self, key: str, model: M, *, expires_after: Optional[int] = None
    ) -> None: ...

    @abc.abstractmethod
    def get(self, key: str) -> Optional[M]: ...

    @abc.abstractmethod
    def delete(self, key: str) -> None: ...

    @abc.abstractmethod
    def refresh(self, key: str, *, expires_after: Optional[int] = None) -> None: ...


class InMemoryKeyValueStore(KeyValueStore[M]):
    def __init__(self) -> None:
        self._store: dict[str, tuple[M, Optional[float]]] = {}
        self._timer: Callable[[], float] = time.monotonic

    def set(self, key: str, model: M, *, expires_after: Optional[int] = None) -> None:
        if expires_after is not None:
            expires_at = self._timer() + expires_after
        else:
            expires_at = None
        self._store[key] = (model, expires_at)

    def get(self, key: str) -> Optional[M]:
        value = self._store.get(key)
        if value is None:
            return None

        model, expires_at = value
        if expires_at is not None and self._timer() >= expires_at:
            self.delete(key)
            return None

        return model

    def delete(self, key: str) -> None:
        if key not in self._store:
            return

        del self._store[key]

    def refresh(self, key: str, *, expires_after: Optional[int] = None) -> None:
        value = self._store.get(key)
        if value is None:
            return

        model, _ = value
        self.set(key, model, expires_after=expires_after)


class RedisKeyValueStore(KeyValueStore[M]):
    @classmethod
    def requirements(cls) -> list[Requirement]:
        return [PackageRequirement("redis")]

    def __init__(self) -> None:
        import redis

        self._r = redis.Redis(
            host=os.environ.get("RAGNA_REDIS_HOST", "localhost"),
            port=int(os.environ.get("RAGNA_REDIS_PORT", 6379)),
        )

    def set(self, key: str, model: M, *, expires_after: Optional[int] = None) -> None:
        self._r.set(key, self.serialize(model), ex=expires_after)

    def get(self, key: str) -> Optional[M]:
        value = cast(bytes, self._r.get(key))
        if value is None:
            return None
        return self.deserialize(value)

    def delete(self, key: str) -> None:
        self._r.delete(key)

    def refresh(self, key: str, *, expires_after: Optional[int] = None) -> None:
        if expires_after is None:
            self._r.persist(key)
        else:
            self._r.expire(key, expires_after)



import contextlib
import threading
import time
import uuid
import webbrowser
from pathlib import Path
from typing import AsyncContextManager, AsyncIterator, Callable, Optional, cast

import httpx
import panel.io.fastapi
from fastapi import FastAPI, Request, status
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse, Response
from fastapi.staticfiles import StaticFiles

import ragna
from ragna.core import RagnaException

from . import _schemas as schemas
from ._api import make_router as make_api_router
from ._auth import UserDependency
from ._config import Config
from ._engine import Engine
from ._ui import app as make_ui_app
from ._utils import handle_localhost_origins, redirect, set_redirect_root_path


def make_app(
    config: Config,
    *,
    api: bool,
    ui: bool,
    ignore_unavailable_components: bool,
    open_browser: bool,
) -> FastAPI:
    set_redirect_root_path(config.root_path)

    lifespan: Optional[Callable[[FastAPI], AsyncContextManager]]
    if open_browser:

        @contextlib.asynccontextmanager
        async def lifespan(app: FastAPI) -> AsyncIterator[None]:
            def target() -> None:
                client = httpx.Client(base_url=config._url)

                def server_available() -> bool:
                    try:
                        return client.get("/health").is_success
                    except httpx.ConnectError:
                        return False

                while not server_available():
                    time.sleep(0.1)

                webbrowser.open(config._url)

            # We are starting the browser on a thread, because the server can only
            # become available _after_ the yield below. By setting daemon=True, the
            # thread will automatically terminated together with the main thread. This
            # is only relevant when the server never becomes available, e.g. if an error
            # occurs. In this case our thread would be stuck in an endless loop.
            thread = threading.Thread(target=target, daemon=True)
            thread.start()
            yield

    else:
        lifespan = None

    app = FastAPI(title="Ragna", version=ragna.__version__, lifespan=lifespan)

    app.add_middleware(
        CORSMiddleware,
        allow_origins=handle_localhost_origins(config.origins),
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    engine = Engine(
        config=config,
        ignore_unavailable_components=ignore_unavailable_components,
    )

    config.auth._add_to_app(app, config=config, engine=engine, api=api, ui=ui)

    if api:
        app.include_router(make_api_router(engine), prefix="/api")

    if ui:
        ui_app = make_ui_app(engine)
        panel.io.fastapi.add_applications({"/ui": ui_app.index_page}, app=app)
        for dir in ["css", "imgs"]:
            app.mount(
                f"/{dir}",
                StaticFiles(directory=str(Path(__file__).parent / "_ui" / dir)),
                name=dir,
            )

    @app.get("/", include_in_schema=False)
    async def base_redirect() -> Response:
        return redirect("/ui" if ui else "/docs")

    @app.get("/health")
    async def health() -> Response:
        return Response(b"", status_code=status.HTTP_200_OK)

    @app.get("/version")
    async def version() -> str:
        return ragna.__version__

    @app.get("/user")
    async def user(user: UserDependency) -> schemas.User:
        return user

    @app.get("/api-keys")
    def list_api_keys(user: UserDependency) -> list[schemas.ApiKey]:
        return engine.list_api_keys(user=user.name)

    @app.post("/api-keys")
    def create_api_key(
        user: UserDependency, api_key_creation: schemas.ApiKeyCreation
    ) -> schemas.ApiKey:
        return engine.create_api_key(user=user.name, api_key_creation=api_key_creation)

    @app.delete("/api-keys/{id}")
    def delete_api_key(user: UserDependency, id: uuid.UUID) -> None:
        return engine.delete_api_key(user=user.name, id=id)

    @app.exception_handler(RagnaException)
    async def ragna_exception_handler(
        request: Request, exc: RagnaException
    ) -> JSONResponse:
        if exc.http_detail is RagnaException.EVENT:
            detail = exc.event
        elif exc.http_detail is RagnaException.MESSAGE:
            detail = str(exc)
        else:
            detail = cast(str, exc.http_detail)
        return JSONResponse(
            status_code=exc.http_status_code,
            content={"error": {"message": detail}},
        )

    return app



from datetime import datetime

import panel as pn
import param

from ragna import __version__ as ragna_version


class LeftSidebar(pn.viewable.Viewer):
    chats = param.List(default=[])
    current_chat_id = param.String(default=None)
    refresh_counter = param.Integer(default=0)

    def __init__(self, api_wrapper, **params):
        super().__init__(**params)

        self.api_wrapper = api_wrapper
        self.on_click_chat = None
        self.on_click_new_chat = None

        self.chat_buttons = []

        pn.state.location.sync(
            self,
            {"current_chat_id": "current_chat_id"},
            on_error=lambda x: print(f"error sync on {x}"),
        )

    def trigger_on_click_new_chat(self, event):
        if event.old > event.new:
            return

        if self.on_click_new_chat is not None:
            self.on_click_new_chat(event)

    def on_click_chat_wrapper(self, event, chat):
        # This is a hack to avoid the event being triggered twice in a row
        if event.old > event.new:
            return

        # update the UI, unselect all buttons ...
        for button in self.chat_buttons:
            if "selected" in button.css_classes:
                button.css_classes = ["chat_button"]

        # ... and select the one that was clicked
        event.obj.css_classes = ["chat_button", "selected"]

        # call the actual callback
        if self.on_click_chat is not None:
            self.on_click_chat(chat)

    def footer(self):
        return pn.pane.HTML(
            """<div class='user_block'> Username </div>
                            """,
            css_classes=["left_sidebar_footer"],
        )

    def refresh(self):
        self.refresh_counter += 1

    @pn.depends("refresh_counter", "chats", "current_chat_id", on_init=True)
    def __panel__(self):
        epoch = datetime(1970, 1, 1)
        self.chats.sort(
            key=lambda chat: (
                epoch if not chat["messages"] else chat["messages"][-1]["timestamp"]
            ),
            reverse=True,
        )

        self.chat_buttons = []
        for chat in self.chats:
            button = pn.widgets.Button(
                name=chat["name"],
                css_classes=["chat_button"],
            )
            button.on_click(lambda event, c=chat: self.on_click_chat_wrapper(event, c))

            self.chat_buttons.append(button)

            try:
                if chat["id"] == self.current_chat_id:
                    button.css_classes.append("selected")
            except Exception:
                pass

        header = pn.pane.HTML(
            """<img src="imgs/ragna_logo.svg" height="32px" /><span>Ragna</span>""",
            css_classes=["left_sidebar_header"],
        )

        new_chat_button = pn.widgets.Button(
            name="New Chat",
            button_type="primary",
            icon="plus",
            css_classes=["new_chat_button"],
        )

        new_chat_button.on_click(self.trigger_on_click_new_chat)

        objects = (
            [header, new_chat_button]
            + self.chat_buttons
            + [
                pn.layout.VSpacer(),
                pn.pane.HTML(f"user: {self.api_wrapper._user}"),
                pn.pane.HTML(f"version: {ragna_version}"),
                # self.footer()
            ]
        )

        result = pn.Column(
            *objects,
            css_classes=["left_sidebar_main_column"],
        )

        return result



__all__ = [
    "ClaudeHaiku",
    "ClaudeOpus",
    "ClaudeSonnet",
    "Command",
    "CommandLight",
    "GeminiPro",
    "GeminiUltra",
    "OllamaGemma2B",
    "OllamaPhi2",
    "OllamaLlama2",
    "OllamaLlava",
    "OllamaMistral",
    "OllamaMixtral",
    "OllamaOrcaMini",
    "Gpt35Turbo16k",
    "Gpt4",
    "Jurassic2Ultra",
    "LlamafileAssistant",
    "RagnaDemoAssistant",
]

from ._ai21labs import Jurassic2Ultra
from ._anthropic import ClaudeHaiku, ClaudeOpus, ClaudeSonnet
from ._cohere import Command, CommandLight
from ._demo import RagnaDemoAssistant
from ._google import GeminiPro, GeminiUltra
from ._llamafile import LlamafileAssistant
from ._ollama import (
    OllamaGemma2B,
    OllamaLlama2,
    OllamaLlava,
    OllamaMistral,
    OllamaMixtral,
    OllamaOrcaMini,
    OllamaPhi2,
)
from ._openai import Gpt4, Gpt35Turbo16k

# isort: split

from ragna._utils import fix_module

fix_module(globals())
del fix_module



from __future__ import annotations

import functools
from typing import Callable, Literal, Optional, cast

import panel as pn
import param
from panel.reactive import ReactiveHTML

from . import styles as ui


class CopyToClipboardButton(ReactiveHTML):
    title = param.String(default=None, doc="The title of the button ")
    value = param.String(default=None, doc="The text to copy to the clipboard.")

    _template = """
        <div type="button" 
                id="copy-button"
                onclick="${script('copy_to_clipboard')}"
                class="container"
                style="cursor: pointer;"
        >
            <svg xmlns="http://www.w3.org/2000/svg" class="icon icon-tabler icon-tabler-clipboard" width="16" height="16" 
                    viewBox="0 0 24 24" stroke-width="2" stroke="gray" fill="none" stroke-linecap="round" stroke-linejoin="round">
                <path stroke="none" d="M0 0h24v24H0z" fill="none"/>
                <path d="M9 5h-2a2 2 0 0 0 -2 2v12a2 2 0 0 0 2 2h10a2 2 0 0 0 2 -2v-12a2 2 0 0 0 -2 -2h-2" />
                <path d="M9 3m0 2a2 2 0 0 1 2 -2h2a2 2 0 0 1 2 2v0a2 2 0 0 1 -2 2h-2a2 2 0 0 1 -2 -2z" />
            </svg>
            <span>${title}</span>
        </div>            
        """

    _scripts = {
        "copy_to_clipboard": """navigator.clipboard.writeText(`${data.value}`);"""
    }

    _stylesheets = ["css/chat_interface/copybutton.css"]


class RagnaChatMessage(pn.chat.ChatMessage):
    role: str = param.Selector(objects=["system", "user", "assistant"])
    sources = param.List(allow_None=True)
    on_click_source_info_callback = param.Callable(allow_None=True)

    def __init__(
        self,
        content: str,
        *,
        role: Literal["system", "user", "assistant"],
        user: str,
        sources: Optional[list[dict]] = None,
        on_click_source_info_callback: Optional[Callable] = None,
        timestamp=None,
        show_timestamp=True,
        assistant_toolbar_visible=True,  # hide the toolbar during streaming
    ):
        css_class = f"message-content-{self.role}"
        self.content_pane = pn.pane.Markdown(
            content,
            css_classes=["message-content", css_class],
        )

        # we make this available on the instance so that we can update the value later
        self.clipboard_button = CopyToClipboardButton(
            value=self.content_pane.object, title="Copy"
        )

        # we make this available on the instance so that we can toggle the visibility
        self.assistant_toolbar = pn.Row(
            self.clipboard_button,
            pn.widgets.Button(
                name="Source Info",
                icon="info-circle",
                css_classes=["source-info-button"],
                on_click=lambda event: self.on_click_source_info_callback(
                    event, self.sources
                ),
            ),
            visible=assistant_toolbar_visible,
        )

        if role == "assistant":
            assert sources is not None
            object = pn.Column(
                self.content_pane,
                self.assistant_toolbar,
                css_classes=["message-content-assistant-with-buttons"],
            )
        else:
            object = self.content_pane

        object.css_classes.append(
            "message-content-no-border" if role == "user" else "message-content-border"
        )

        super().__init__(
            object=object,
            role=role,
            user=user,
            sources=sources,
            on_click_source_info_callback=on_click_source_info_callback,
            timestamp=timestamp,
            show_timestamp=show_timestamp,
            show_reaction_icons=False,
            show_user=False,
            show_copy_icon=False,
            css_classes=[f"message-{role}"],
            avatar_lookup=functools.partial(self._ragna_avatar_lookup, role=role),
        )
        self._stylesheets.append("css/chat_interface/chatmessage.css")

    # This cannot be a bound method, because it creates a reference cycle when trying
    # to access the repr of the message. See
    # https://github.com/Quansight/ragna/issues/359 for details.
    @staticmethod
    def _ragna_avatar_lookup(user: str, *, role: str) -> str:
        if role == "system":
            return "imgs/ragna_logo.svg"
        elif role == "user":
            return "👤"

        try:
            organization, model = user.split("/")
        except ValueError:
            organization = ""
            model = user

        if organization == "Ragna":
            return "imgs/ragna_logo.svg"
        elif organization == "OpenAI":
            if model.startswith("gpt-3"):
                return "https://upload.wikimedia.org/wikipedia/commons/thumb/0/04/ChatGPT_logo.svg/1024px-ChatGPT_logo.svg.png?20230318122128"
            elif model.startswith("gpt-4"):
                return "https://upload.wikimedia.org/wikipedia/commons/a/a4/GPT-4.png"
        elif organization == "Anthropic":
            return "https://upload.wikimedia.org/wikipedia/commons/1/14/Anthropic.png"

        return model[0].upper()


class RagnaChatInterface(pn.chat.ChatInterface):
    get_user_from_role = param.Callable(allow_None=True)

    @param.depends("placeholder_text", watch=True, on_init=True)
    def _update_placeholder(self):
        self._placeholder = RagnaChatMessage(
            ui.message_loading_indicator,
            role="system",
            user=self.get_user_from_role("system"),
            show_timestamp=False,
        )

    def _build_message(self, *args, **kwargs) -> Optional[RagnaChatMessage]:
        message = super()._build_message(*args, **kwargs)
        if message is None:
            return None

        # We only ever hit this function for user inputs, since we control the
        # generation of the system and assistant messages manually. Thus, we can
        # unconditionally create a user message here.
        return RagnaChatMessage(message.object, role="user", user=self.user)


class CentralView(pn.viewable.Viewer):
    current_chat = param.ClassSelector(class_=dict, default=None)

    def __init__(self, api_wrapper, **params):
        super().__init__(**params)

        # FIXME: make this dynamic from the login
        self.user = ""
        self.api_wrapper = api_wrapper
        self.chat_info_button = pn.widgets.Button(
            # The name will be filled at runtime in self.header
            name="",
            on_click=self.on_click_chat_info_wrapper,
            icon="info-circle",
            css_classes=["chat-info-button"],
        )
        self.on_click_chat_info = None

    def on_click_chat_info_wrapper(self, event):
        if self.on_click_chat_info is None:
            return

        pills = "".join(
            [
                f"""<div class='chat_document_pill'>{d['name']}</div>"""
                for d in self.current_chat["documents"]
            ]
        )

        grid_height = len(self.current_chat["documents"]) // 3

        markdown = "\n".join(
            [
                "To change configurations, start a new chat.\n",
                "**Uploaded Files**",
                f"<div class='pills_list'>{pills}</div><br />\n\n",
                "----",
                "**Source Storage**",
                f"""<span>{self.current_chat['source_storage']}</span>\n""",
                "----",
                "**Assistant**",
                f"""<span>{self.current_chat['assistant']}</span>\n""",
                "**Advanced configuration**",
                *[
                    f"- **{key.replace('_', ' ').title()}**: {value}"
                    for key, value in self.current_chat["params"].items()
                ],
            ]
        )

        self.on_click_chat_info(
            event,
            "Chat Info",
            [
                pn.pane.Markdown(
                    markdown,
                    dedent=True,
                    css_classes=["chat_info_markdown"],
                    # The CSS rule below relies on a variable value, so we can't move it into modifers
                    stylesheets=[
                        ui.css(
                            ":host(.chat_info_markdown) .pills_list",
                            {
                                "grid-template": f"repeat({grid_height}, 1fr) / repeat(3, 1fr)",
                            },
                        )
                    ],
                ),
            ],
        )

    def on_click_source_info_wrapper(self, event, sources):
        if self.on_click_chat_info is None:
            return

        source_infos = []
        for rank, source in enumerate(sources, 1):
            location = source["location"]
            if location:
                location = f": page(s) {location}"
            source_infos.append(
                (
                    f"<b>{rank}. {source['document']['name']}</b> {location}",
                    pn.pane.Markdown(source["content"], css_classes=["source-content"]),
                )
            )

        self.on_click_chat_info(
            event,
            "Source Info",
            [
                pn.pane.Markdown(
                    "This response was generated using the following data from the uploaded files: <br />",
                ),
                pn.layout.Accordion(
                    *source_infos,
                    header_background="transparent",
                    css_classes=["source-accordion"],
                ),
            ],
        )

    def set_current_chat(self, chat):
        self.current_chat = chat

    def get_user_from_role(self, role: Literal["system", "user", "assistant"]) -> str:
        if role == "system":
            return "Ragna"
        elif role == "user":
            return cast(str, self.user)
        elif role == "assistant":
            return cast(str, self.current_chat["assistant"])
        else:
            raise RuntimeError

    async def chat_callback(
        self, content: str, user: str, instance: pn.chat.ChatInterface
    ):
        try:
            answer_stream = self.api_wrapper.answer(self.current_chat["id"], content)
            answer = await anext(answer_stream)

            message = RagnaChatMessage(
                answer["content"],
                role="assistant",
                user=self.get_user_from_role("assistant"),
                sources=answer["sources"],
                on_click_source_info_callback=self.on_click_source_info_wrapper,
                assistant_toolbar_visible=False,
            )
            yield message

            async for chunk in answer_stream:
                message.content_pane.object += chunk["content"]
            message.clipboard_button.value = message.content_pane.object
            message.assistant_toolbar.visible = True

        except Exception:
            yield RagnaChatMessage(
                (
                    "Sorry, something went wrong. "
                    "If this problem persists, please contact your administrator."
                ),
                role="system",
                user=self.get_user_from_role("system"),
            )

    @pn.depends("current_chat")
    def chat_interface(self):
        if self.current_chat is None:
            return

        return RagnaChatInterface(
            *[
                RagnaChatMessage(
                    message["content"],
                    role=message["role"],
                    user=self.get_user_from_role(message["role"]),
                    sources=message["sources"],
                    timestamp=message["timestamp"],
                    on_click_source_info_callback=self.on_click_source_info_wrapper,
                )
                for message in self.current_chat["messages"]
            ],
            callback=self.chat_callback,
            user=self.user,
            get_user_from_role=self.get_user_from_role,
            show_rerun=False,
            show_undo=False,
            show_clear=False,
            show_button_name=False,
            view_latest=True,
            sizing_mode="stretch_width",
            # TODO: Remove the parameter when
            #  https://github.com/holoviz/panel/issues/6115 is merged and released. We
            #  currently need it to avoid sending a message when the text input is
            #  de-focussed. But this also means we can't hit enter to send.
            auto_send_types=[],
            widgets=[
                pn.widgets.TextInput(
                    placeholder="Ask a question about the documents",
                    # css_classes is forced to chat-interface-input-widget by Panel
                )
            ],
            show_activity_dot=False,
        )

    @pn.depends("current_chat")
    def header(self):
        if self.current_chat is None:
            return

        current_chat_name = ""
        if self.current_chat is not None:
            current_chat_name = self.current_chat["name"]

        chat_name_header = pn.pane.HTML(
            f"<p>{current_chat_name}</p>",
            sizing_mode="stretch_width",
            css_classes=["chat-name-header"],
        )

        chat_documents_pills = []
        if self.current_chat is not None:
            doc_names = [d["name"] for d in self.current_chat["documents"]]

            # FIXME: Instead of setting a hard limit of 20 documents here, this should
            #  scale automatically with the width of page
            #  See https://github.com/Quansight/ragna/issues/224
            for doc_name in doc_names[:20]:
                pill = pn.pane.HTML(
                    f"""<div class="chat-document-pill">{doc_name}</div>""",
                    css_classes=["chat-document-pill"],
                )

                chat_documents_pills.append(pill)

        self.chat_info_button.name = (
            f"{self.current_chat['assistant']} | {self.current_chat['source_storage']}"
        )

        return pn.Row(
            chat_name_header,
            *chat_documents_pills,
            self.chat_info_button,
            css_classes=["central-view-header"],
        )

    def set_loading(self, is_loading):
        self.main_column.loading = is_loading

    def __panel__(self):
        self.main_column = pn.Column(
            self.header,
            self.chat_interface,
            sizing_mode="stretch_width",
            css_classes=["central-view-main-column"],
        )

        return self.main_column



import json
from datetime import datetime, timezone
from typing import Any, Optional

from sqlalchemy import Column, ForeignKey, Table, types
from sqlalchemy.engine import Dialect
from sqlalchemy.orm import DeclarativeBase, relationship  # type: ignore[attr-defined]

from ragna.core import MessageRole


class Json(types.TypeDecorator):
    """Universal JSON type which stores values as strings.

    This is needed because sqlalchemy.types.JSON only works for a limited subset of
    databases.
    """

    impl = types.String

    cache_ok = True

    def process_bind_param(self, value: Any, dialect: Dialect) -> str:
        return json.dumps(value)

    def process_result_value(
        self,
        value: str,  # type: ignore[override]
        dialect: Dialect,
    ) -> Any:
        return json.loads(value)


class UtcAwareDateTime(types.TypeDecorator):
    """UTC timezone aware datetime type.

    This is needed because sqlalchemy.types.DateTime(timezone=True) does not
    consistently store the timezone.
    """

    impl = types.DateTime

    cache_ok = True

    def process_bind_param(  # type: ignore[override]
        self, value: Optional[datetime], dialect: Dialect
    ) -> Optional[datetime]:
        if value is not None:
            assert value.tzinfo == timezone.utc

        return value

    def process_result_value(
        self, value: Optional[datetime], dialect: Dialect
    ) -> Optional[datetime]:
        if value is None:
            return None

        return value.replace(tzinfo=timezone.utc)


class Base(DeclarativeBase):
    pass


class User(Base):
    __tablename__ = "users"

    id = Column(types.Uuid, primary_key=True)  # type: ignore[attr-defined]
    name = Column(types.String, nullable=False, unique=True)
    api_keys = relationship("ApiKey", back_populates="user")


class ApiKey(Base):
    __tablename__ = "api_keys"

    id = Column(types.Uuid, primary_key=True)  # type: ignore[attr-defined]

    user_id = Column(ForeignKey("users.id"))
    user = relationship("User", back_populates="api_keys")

    name = Column(types.String, nullable=False)
    value = Column(types.String, nullable=False, unique=True)
    expires_at = Column(UtcAwareDateTime, nullable=False)


document_chat_association_table = Table(
    "document_chat_association_table",
    Base.metadata,
    Column("document_id", ForeignKey("documents.id"), primary_key=True),
    Column("chat_id", ForeignKey("chats.id"), primary_key=True),
)


class Document(Base):
    __tablename__ = "documents"

    id = Column(types.Uuid, primary_key=True)  # type: ignore[attr-defined]
    user_id = Column(ForeignKey("users.id"))
    name = Column(types.String, nullable=False)
    # Mind the trailing underscore here. Unfortunately, this is necessary, because
    # metadata without the underscore is reserved by SQLAlchemy
    metadata_ = Column(Json, nullable=False)
    chats = relationship(
        "Chat",
        secondary=document_chat_association_table,
        back_populates="documents",
    )
    sources = relationship(
        "Source",
        back_populates="document",
    )


class Chat(Base):
    __tablename__ = "chats"

    id = Column(types.Uuid, primary_key=True)  # type: ignore[attr-defined]
    user_id = Column(ForeignKey("users.id"))
    name = Column(types.String, nullable=False)
    documents = relationship(
        "Document",
        secondary=document_chat_association_table,
        back_populates="chats",
    )
    source_storage = Column(types.String, nullable=False)
    assistant = Column(types.String, nullable=False)
    params = Column(Json, nullable=False)
    messages = relationship(
        "Message", cascade="all, delete", order_by="Message.timestamp"
    )
    prepared = Column(types.Boolean, nullable=False)


source_message_association_table = Table(
    "source_message_association_table",
    Base.metadata,
    Column("source_id", ForeignKey("sources.id"), primary_key=True),
    Column("message_id", ForeignKey("messages.id"), primary_key=True),
)


class Source(Base):
    __tablename__ = "sources"

    # This is not a UUID column as all of the other IDs, because we don't control its
    # generation. It is generated as part of ragna.core.SourceStorage.retrieve and using
    # a string here doesn't impose unnecessary constraints on the user.
    id = Column(types.String, primary_key=True)

    document_id = Column(ForeignKey("documents.id"))
    document = relationship("Document", back_populates="sources")

    location = Column(types.String, nullable=False)
    content = Column(types.String, nullable=False)
    num_tokens = Column(types.Integer, nullable=False)

    messages = relationship(
        "Message",
        secondary=source_message_association_table,
        back_populates="sources",
    )


class Message(Base):
    __tablename__ = "messages"

    id = Column(types.Uuid, primary_key=True)  # type: ignore[attr-defined]
    chat_id = Column(ForeignKey("chats.id"))
    content = Column(types.String, nullable=False)
    role = Column(types.Enum(MessageRole), nullable=False)
    sources = relationship(
        "Source",
        secondary=source_message_association_table,
        back_populates="messages",
    )

    timestamp = Column(UtcAwareDateTime, nullable=False)



import asyncio
import itertools
import json
import os
from pathlib import Path

import httpx
import pytest

from ragna import assistants
from ragna._utils import BackgroundSubprocess
from ragna.assistants._http_api import HttpApiAssistant, HttpStreamingProtocol
from ragna.core import Message, RagnaException
from tests.utils import get_available_port, skip_on_windows

HTTP_API_ASSISTANTS = [
    assistant
    for assistant in assistants.__dict__.values()
    if isinstance(assistant, type)
    and issubclass(assistant, HttpApiAssistant)
    and assistant is not HttpApiAssistant
]


@skip_on_windows
@pytest.mark.parametrize(
    "assistant",
    [assistant for assistant in HTTP_API_ASSISTANTS if assistant._API_KEY_ENV_VAR],
)
async def test_api_call_error_smoke(mocker, assistant):
    mocker.patch.dict(os.environ, {assistant._API_KEY_ENV_VAR: "SENTINEL"})

    messages = [Message(content="?", sources=[])]
    chunks = assistant().answer(messages)

    with pytest.raises(RagnaException, match="API call failed"):
        await anext(chunks)


@pytest.fixture
def streaming_server():
    port = get_available_port()
    base_url = f"http://localhost:{port}"

    def check_fn():
        try:
            return httpx.get(f"{base_url}/health").is_success
        except httpx.ConnectError:
            return False

    with BackgroundSubprocess(
        "uvicorn",
        f"--app-dir={Path(__file__).parent}",
        f"--port={port}",
        "streaming_server:app",
        startup_fn=check_fn,
    ):
        yield base_url


class HttpStreamingAssistant(HttpApiAssistant):
    _API_KEY_ENV_VAR = None

    @staticmethod
    def new(base_url, streaming_protocol):
        cls = type(
            f"{streaming_protocol.name.title()}{HttpStreamingAssistant.__name__}",
            (HttpStreamingAssistant,),
            dict(_STREAMING_PROTOCOL=streaming_protocol),
        )
        return cls(base_url)

    def __init__(self, base_url):
        super().__init__()
        self._endpoint = f"{base_url}/{self._STREAMING_PROTOCOL.name.lower()}"

    async def answer(self, messages):
        if self._STREAMING_PROTOCOL is HttpStreamingProtocol.JSON:
            parse_kwargs = dict(item="item")
        else:
            parse_kwargs = dict()

        async with self._call_api(
            "POST",
            self._endpoint,
            content=messages[-1].content,
            parse_kwargs=parse_kwargs,
        ) as stream:
            async for chunk in stream:
                if chunk.get("break"):
                    break

                yield chunk


@skip_on_windows
@pytest.mark.parametrize("streaming_protocol", list(HttpStreamingProtocol))
async def test_http_streaming(streaming_server, streaming_protocol):
    assistant = HttpStreamingAssistant.new(streaming_server, streaming_protocol)

    data = [{"chunk": chunk} for chunk in ["foo", "bar", "baz"]]
    expected_chunks = iter(data)
    actual_chunks = assistant.answer([Message(content=json.dumps(data))])
    async for actual_chunk in actual_chunks:
        expected_chunk = next(expected_chunks)
        assert actual_chunk == expected_chunk

    with pytest.raises(StopIteration):
        next(expected_chunks)


@skip_on_windows
@pytest.mark.parametrize("streaming_protocol", list(HttpStreamingProtocol))
def test_http_streaming_termination(streaming_server, streaming_protocol):
    # Non-regression test for https://github.com/Quansight/ragna/pull/462

    async def main():
        assistant = HttpStreamingAssistant.new(streaming_server, streaming_protocol)

        data = [
            {"chunk": "foo", "break": False},
            {"chunk": "bar", "break": False},
            {"chunk": "baz", "break": True},
        ]
        expected_chunks = itertools.takewhile(lambda chunk: not chunk["break"], data)
        actual_chunks = assistant.answer([Message(content=json.dumps(data))])
        async for actual_chunk in actual_chunks:
            expected_chunk = next(expected_chunks)
            assert actual_chunk == expected_chunk

        with pytest.raises(StopIteration):
            next(expected_chunks)

    asyncio.run(main())



import os
from functools import cached_property

from ._http_api import HttpStreamingProtocol
from ._openai import OpenaiLikeHttpApiAssistant


class LlamafileAssistant(OpenaiLikeHttpApiAssistant):
    """[llamafile](https://github.com/Mozilla-Ocho/llamafile)

    To use this assistant, start the llamafile server manually. By default, the server
    is expected at `http://localhost:8080`. This can be changed with the
    `RAGNA_LLAMAFILE_BASE_URL` environment variable.

    !!! info "Required packages"

        - `httpx_sse`
    """

    _API_KEY_ENV_VAR = None
    _STREAMING_PROTOCOL = HttpStreamingProtocol.SSE
    _MODEL = None

    @classmethod
    def display_name(cls) -> str:
        return "llamafile"

    @cached_property
    def _url(self) -> str:
        base_url = os.environ.get("RAGNA_LLAMAFILE_BASE_URL", "http://localhost:8080")
        return f"{base_url}/v1/chat/completions"



import abc
from functools import cached_property
from typing import Any, AsyncContextManager, AsyncIterator, Optional, cast

from ragna.core import Message, Source

from ._http_api import HttpApiAssistant, HttpStreamingProtocol


class OpenaiLikeHttpApiAssistant(HttpApiAssistant):
    _MODEL: Optional[str]

    @property
    @abc.abstractmethod
    def _url(self) -> str: ...

    def _make_system_content(self, sources: list[Source]) -> str:
        # See https://github.com/openai/openai-cookbook/blob/main/examples/How_to_format_inputs_to_ChatGPT_models.ipynb
        instruction = (
            "You are an helpful assistants that answers user questions given the context below. "
            "If you don't know the answer, just say so. Don't try to make up an answer. "
            "Only use the sources below to generate the answer."
        )
        return instruction + "\n\n".join(source.content for source in sources)

    def _call_openai_api(
        self, prompt: str, sources: list[Source], *, max_new_tokens: int
    ) -> AsyncContextManager[AsyncIterator[dict[str, Any]]]:
        # See https://platform.openai.com/docs/api-reference/chat/create
        # and https://platform.openai.com/docs/api-reference/chat/streaming
        headers = {
            "Content-Type": "application/json",
        }
        if self._api_key is not None:
            headers["Authorization"] = f"Bearer {self._api_key}"

        json_ = {
            "messages": [
                {
                    "role": "system",
                    "content": self._make_system_content(sources),
                },
                {
                    "role": "user",
                    "content": prompt,
                },
            ],
            "temperature": 0.0,
            "max_tokens": max_new_tokens,
            "stream": True,
        }
        if self._MODEL is not None:
            json_["model"] = self._MODEL

        return self._call_api("POST", self._url, headers=headers, json=json_)

    async def answer(
        self, messages: list[Message], *, max_new_tokens: int = 256
    ) -> AsyncIterator[str]:
        prompt, sources = (message := messages[-1]).content, message.sources
        async with self._call_openai_api(
            prompt, sources, max_new_tokens=max_new_tokens
        ) as stream:
            async for data in stream:
                choice = data["choices"][0]
                if choice["finish_reason"] is not None:
                    break

                yield cast(str, choice["delta"]["content"])


class OpenaiAssistant(OpenaiLikeHttpApiAssistant):
    _API_KEY_ENV_VAR = "OPENAI_API_KEY"
    _STREAMING_PROTOCOL = HttpStreamingProtocol.SSE

    @classmethod
    def display_name(cls) -> str:
        return f"OpenAI/{cls._MODEL}"

    @cached_property
    def _url(self) -> str:
        return "https://api.openai.com/v1/chat/completions"


class Gpt35Turbo16k(OpenaiAssistant):
    """[OpenAI GPT-3.5](https://platform.openai.com/docs/models/gpt-3-5)

    !!! info "Required environment variables"

        - `OPENAI_API_KEY`

    !!! info "Required packages"

        - `httpx_sse`
    """

    _MODEL = "gpt-3.5-turbo-16k"


class Gpt4(OpenaiAssistant):
    """[OpenAI GPT-4](https://platform.openai.com/docs/models/gpt-4)

    !!! info "Required environment variables"

        - `OPENAI_API_KEY`

    !!! info "Required packages"

        - `httpx_sse`
    """

    _MODEL = "gpt-4"



import atexit
import inspect
import itertools
import os
import subprocess
import sys
import tempfile
import textwrap
from pathlib import Path
from typing import Any, Optional, cast

import httpx

from ragna.core import RagnaException
from ragna.deploy import Config

from ._utils import BackgroundSubprocess

__all__ = ["SAMPLE_CONTENT", "RagnaDeploy"]

SAMPLE_CONTENT = """\
Ragna is an open source project built by Quansight. It is designed to allow
organizations to explore the power of Retrieval-augmented generation (RAG) based
AI tools. Ragna provides an intuitive API for quick experimentation and built-in
tools for creating production-ready applications allowing you to quickly leverage
Large Language Models (LLMs) for your work.

The Ragna website is https://ragna.chat/. The source code is available at
https://github.com/Quansight/ragna under the BSD 3-Clause license.
"""


class RagnaDeploy:
    def __init__(self, config: Config) -> None:
        self.config = config
        python_path, config_path = self._prepare_config()
        self._process = self._deploy(config, config_path, python_path)
        # In case the documentation errors before we call RagnaDeploy.terminate,
        # we still need to stop the server to avoid zombie processes
        atexit.register(self.terminate, quiet=True)

    def _prepare_config(self) -> tuple[str, str]:
        deploy_directory = Path(tempfile.mkdtemp())

        python_path = os.pathsep.join(
            [str(deploy_directory), os.environ.get("PYTHONPATH", "")]
        )
        config_path = str(deploy_directory / "ragna.toml")

        self.config.local_root = deploy_directory
        self.config.database_url = f"sqlite:///{deploy_directory / 'ragna.db'}"

        sys.modules["__main__"].__file__ = inspect.getouterframes(
            inspect.currentframe()
        )[2].filename
        custom_module = deploy_directory.name
        custom_components = set()
        with open(deploy_directory / f"{custom_module}.py", "w") as file:
            # FIXME Find a way to automatically detect necessary imports
            file.write("import uuid; from uuid import *\n")
            file.write("import textwrap; from textwrap import*\n")
            file.write("from typing import *\n")
            file.write("from ragna import *\n")
            file.write("from ragna.core import *\n")

            for component in itertools.chain(
                self.config.source_storages, self.config.assistants
            ):
                if component.__module__ == "__main__":
                    custom_components.add(component)
                    file.write(f"{textwrap.dedent(inspect.getsource(component))}\n\n")
                    component.__module__ = custom_module

        self.config.to_file(config_path)

        for component in custom_components:
            component.__module__ = "__main__"

        return python_path, config_path

    def _deploy(
        self, config: Config, config_path: str, python_path: str
    ) -> BackgroundSubprocess:
        env = os.environ.copy()
        env["PYTHONPATH"] = python_path

        def startup_fn() -> bool:
            try:
                return httpx.get(f"{config._url}/health").is_success
            except httpx.ConnectError:
                return False

        if startup_fn():
            raise RagnaException("ragna server is already running")

        return BackgroundSubprocess(
            sys.executable,
            "-m",
            "ragna",
            "deploy",
            "--api",
            "--no-ui",
            "--config",
            config_path,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            env=env,
            startup_fn=startup_fn,
            startup_timeout=60,
        )

    def get_http_client(
        self,
        *,
        authenticate: bool = False,
        upload_sample_document: bool = False,
    ) -> tuple[httpx.Client, Optional[dict[str, Any]]]:
        if upload_sample_document and not authenticate:
            raise RagnaException(
                "Cannot upload a document without authenticating first. "
                "Set authenticate=True when using upload_sample_document=True."
            )

        client = httpx.Client(base_url=self.config._url)

        if authenticate:
            client.get("/login", follow_redirects=True)

        if upload_sample_document:
            name, content = "ragna.txt", SAMPLE_CONTENT

            response = client.post(
                "/api/documents", json=[{"name": name}]
            ).raise_for_status()
            document = cast(dict[str, Any], response.json()[0])

            client.put(
                "/api/documents",
                files=[("documents", (document["id"], content.encode()))],
            )
        else:
            document = None

        return client, document

    def terminate(self, quiet: bool = False) -> None:
        if self._process is None:
            return

        output = self._process.terminate()

        if output and not quiet:
            stdout, _ = output
            print(stdout)



from typing import Optional
from urllib.parse import SplitResult, urlsplit, urlunsplit

from fastapi import status
from fastapi.responses import RedirectResponse

from ragna.core import RagnaException

_REDIRECT_ROOT_PATH: Optional[str] = None


def set_redirect_root_path(root_path: str) -> None:
    global _REDIRECT_ROOT_PATH
    _REDIRECT_ROOT_PATH = root_path


def redirect(
    url: str, *, status_code: int = status.HTTP_303_SEE_OTHER
) -> RedirectResponse:
    if _REDIRECT_ROOT_PATH is None:
        raise RagnaException

    if url.startswith("/"):
        url = _REDIRECT_ROOT_PATH + url

    return RedirectResponse(url, status_code=status_code)


def handle_localhost_origins(origins: list[str]) -> list[str]:
    # Since localhost is an alias for 127.0.0.1, we allow both so users and developers
    # don't need to worry about it.
    localhost_origins = {
        components.hostname: components
        for url in origins
        if (components := urlsplit(url)).hostname in {"127.0.0.1", "localhost"}
    }
    if "127.0.0.1" in localhost_origins and "localhost" not in localhost_origins:
        origins.append(
            urlunsplit(_replace_hostname(localhost_origins["127.0.0.1"], "localhost"))
        )
    elif "localhost" in localhost_origins and "127.0.0.1" not in localhost_origins:
        origins.append(
            urlunsplit(_replace_hostname(localhost_origins["localhost"], "127.0.0.1"))
        )

    return origins


def _replace_hostname(split_result: SplitResult, hostname: str) -> SplitResult:
    # This is a separate function, since hostname is not an element of the SplitResult
    # namedtuple, but only a property. Thus, we need to replace the netloc item, from
    # which the hostname is generated.
    if split_result.port is None:
        netloc = hostname
    else:
        netloc = f"{hostname}:{split_result.port}"
    return split_result._replace(netloc=netloc)
